# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
"""
Provides a modal dialog box for modifying the AWS Deadline Cloud
local workstation configuration.

Example code:
    from deadline.client.ui.dialogs import DeadlineConfigDialog
    DeadlineConfigDialog.configure_settings(parent=self)
"""

__all__ = ["DeadlineConfigDialog"]

import sys
import threading
from configparser import ConfigParser
from logging import getLogger, root
from typing import Callable, Dict, List, Optional

import boto3  # type: ignore[import]
from botocore.exceptions import ProfileNotFound  # type: ignore[import]
from deadline.job_attachments.models import FileConflictResolution, JobAttachmentsFileSystem
from qtpy.QtCore import QSize, Qt, Signal
from qtpy.QtWidgets import (  # pylint: disable=import-error; type: ignore
    QApplication,
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QStyle,
    QVBoxLayout,
    QWidget,
    QScrollArea,
)

import os

from ... import api
from ..deadline_authentication_status import DeadlineAuthenticationStatus
from ...config import config_file, get_setting_default, str2bool
from .. import CancelationFlag, block_signals
from ..widgets import DirectoryPickerWidget
from ..widgets.deadline_authentication_status_widget import DeadlineAuthenticationStatusWidget
from .deadline_login_dialog import DeadlineLoginDialog

logger = getLogger(__name__)

NOT_VALID_MARKER = "[NOT VALID]"


class DeadlineConfigDialog(QDialog):
    """
    A modal dialog box for modifying the AWS Deadline Cloud local workstation
    configuration.

    Example code:
        DeadlineConfigDialog.configure_settings(parent=self)
    """

    @staticmethod
    def configure_settings(parent=None, set_profile_focus=False) -> bool:
        """
        Static method that runs the Deadline Config Dialog.

        Args:
            parent: Parent widget
            set_profile_focus: Optional boolean to set the initial focus to the profile selector

        Returns True if any changes were applied, False otherwise.
        """
        deadline_config = DeadlineConfigDialog(parent=parent)

        if set_profile_focus:
            deadline_config.config_box.aws_profiles_box.setFocus()

        deadline_config.exec_()
        return deadline_config.changes_were_applied

    def __init__(self, parent=None) -> None:
        super().__init__(
            parent=parent, f=Qt.WindowSystemMenuHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint
        )

        self.setWindowTitle("AWS Deadline Cloud workstation configuration")
        self.deadline_authentication_status = DeadlineAuthenticationStatus.getInstance()
        self._build_ui()

    def sizeHint(self):
        # Get available screen space for adaptive sizing
        screen = QApplication.primaryScreen().availableGeometry()
        available_height = screen.height()

        # Calculate optimal dialog height based on screen size
        content_height = 850  # Height needed to show all content without scroll bars
        max_dialog_height = int(available_height * 0.9)  # 90% of screen height

        # Use smaller of content height or max screen percentage
        optimal_height = min(content_height, max_dialog_height)

        return QSize(650, optimal_height)

    def _build_ui(self):
        self.layout = QVBoxLayout(self)

        self.config_box = DeadlineWorkstationConfigWidget(parent=self)

        self.scrollArea = DeadlineScrollArea(self)
        self.scrollArea.setWidget(self.config_box)
        # Enable widget resizing within scroll area
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        # Hide the scroll area border for a cleaner appearance
        self.scrollArea.setStyleSheet("QScrollArea { border: none; }")

        self.layout.addWidget(self.scrollArea)

        self.config_box.refreshed.connect(self.on_refresh)

        self.auth_status_box = DeadlineAuthenticationStatusWidget(
            parent=self, show_profile_switch=False
        )
        self.layout.addWidget(self.auth_status_box)
        self.deadline_authentication_status.deadline_config_changed.connect(self.config_box.refresh)
        self.deadline_authentication_status.api_availability_changed.connect(
            self.on_auth_status_update
        )

        # We only use a Close button, not OK/Cancel, because we live update the settings.
        self.button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel | QDialogButtonBox.Apply, Qt.Horizontal
        )
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        self.button_box.clicked.connect(self.on_button_box_clicked)
        self.auth_status_box.logout_clicked.connect(self.on_logout)
        self.auth_status_box.login_clicked.connect(self.on_login)
        self.layout.addWidget(self.button_box)

        # Refresh the lists so queue/farm show the description instead of the ID
        self.config_box.refresh_lists()

    @property
    def changes_were_applied(self) -> bool:
        return self.config_box.changes_were_applied

    def accept(self):
        if self.config_box.apply():
            super().accept()

    def reject(self):
        self.deadline_authentication_status.set_config(config_file.read_config())
        super().reject()

    def on_login(self):
        DeadlineLoginDialog.login(parent=self, config=self.config_box.config)
        self.deadline_authentication_status.refresh_status()
        self.config_box.refresh()

    def on_logout(self):
        api.logout(config=self.config_box.config)
        self.deadline_authentication_status.refresh_status()
        self.config_box.refresh()

    def on_button_box_clicked(self, button):
        if self.button_box.standardButton(button) == QDialogButtonBox.Apply:
            self.config_box.apply()

    def on_refresh(self):
        # Enable the "Apply" button only if there are changes
        self.button_box.button(QDialogButtonBox.Apply).setEnabled(bool(self.config_box.changes))
        # Update the auth status with the refreshed config
        self.deadline_authentication_status.set_config(self.config_box.config)

    def on_auth_status_update(self):
        # If the AWS Deadline Cloud API is authorized successfully for the AWS profile
        # in the config dialog, refresh the farm/queue lists
        if self.deadline_authentication_status.api_availability and config_file.get_setting(
            "defaults.aws_profile_name", self.deadline_authentication_status.config
        ) == config_file.get_setting("defaults.aws_profile_name", self.config_box.config):
            self.config_box.refresh_lists()


class DeadlineScrollArea(QScrollArea):
    def __init__(self, parent=None):
        super().__init__(parent)

    def sizeHint(self):
        return QSize(500, 400)


class DeadlineWorkstationConfigWidget(QWidget):
    """
    A widget that displays and edits the AWS Deadline Cloud local workstation.
    """

    # Signal for when the GUI is refreshed
    refreshed = Signal()

    # Emitted when an async refresh_queues_list thread completes,
    # provides (aws_profile_name, farm_id, [(queue_id, queue_name), ...])
    _queue_list_update = Signal(str, str, list)
    # Emitted when an async refresh_storage_profiles_name_list thread completes,
    # provides (aws_profile_name, farm_id, queue_id, [storage_profile_id, ...])
    _storage_profile_list_update = Signal(str, str, list)
    # This signal is sent when any background refresh thread catches an exception,
    # provides (operation_name, BaseException)
    _background_exception = Signal(str, BaseException)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.changes = {}
        self.config: Optional[ConfigParser] = None
        self.changes_were_applied = False

        self._build_ui()
        self._fill_aws_profiles_box()
        self.refresh()

    def minimumSizeHint(self):
        return QSize(500, 700)

    def _build_ui(self):
        # Ensure the widget expands horizontally
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        self.v_layout = QVBoxLayout(self)
        self.setLayout(self.v_layout)

        # Set consistent spacing between group boxes and margins
        self.v_layout.setSpacing(20)  # 20px spacing between group boxes
        self.v_layout.setContentsMargins(10, 10, 10, 10)  # 10px margins around content

        # Simplified stylesheet - focus only on title positioning, use layout for spacing
        GROUP_BOX_STYLE_SHEET = """
        QGroupBox::title {
            subcontrol-origin: margin;
            padding-left: 5px;
            top: 0px;
        }
        QGroupBox {
            margin-top: 20px;
        }
        """

        self.labels = {}
        self._refresh_callbacks: List[Callable] = []

        # Global settings
        self.global_settings_group = QGroupBox(parent=self, title="Global settings")
        self.global_settings_group.setStyleSheet(GROUP_BOX_STYLE_SHEET)
        self.global_settings_group.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.v_layout.addWidget(self.global_settings_group)
        global_settings_layout = QFormLayout(self.global_settings_group)
        self._build_global_settings_ui(self.global_settings_group, global_settings_layout)

        # AWS Profile-specific settings
        self.profile_settings_group = QGroupBox(parent=self, title="Profile settings")
        self.profile_settings_group.setStyleSheet(GROUP_BOX_STYLE_SHEET)
        self.profile_settings_group.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.v_layout.addWidget(self.profile_settings_group)
        profile_settings_layout = QFormLayout(self.profile_settings_group)
        self._build_profile_settings_ui(self.profile_settings_group, profile_settings_layout)

        # Farm-specific settings
        self.farm_settings_group = QGroupBox(parent=self, title="Farm settings")
        self.farm_settings_group.setStyleSheet(GROUP_BOX_STYLE_SHEET)
        self.farm_settings_group.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.v_layout.addWidget(self.farm_settings_group)
        farm_settings_layout = QFormLayout(self.farm_settings_group)
        self._build_farm_settings_ui(self.farm_settings_group, farm_settings_layout)

        # General settings
        self.general_settings_group = QGroupBox(parent=self, title="General settings")
        self.general_settings_group.setStyleSheet(GROUP_BOX_STYLE_SHEET)
        self.general_settings_group.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.v_layout.addWidget(self.general_settings_group)
        general_settings_layout = QFormLayout(self.general_settings_group)
        self._build_general_settings_ui(self.general_settings_group, general_settings_layout)

        # Add vertical spacer to push content to top and prevent group boxes from expanding
        self.v_layout.addItem(QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding))

        self._background_exception.connect(self.handle_background_exception)

    def _build_global_settings_ui(self, group, layout):
        layout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)

        self.aws_profiles_box = QComboBox(parent=group)
        aws_profile_label = self.labels["defaults.aws_profile_name"] = QLabel("AWS profile")
        layout.addRow(aws_profile_label, self.aws_profiles_box)
        self.aws_profiles_box.currentTextChanged.connect(self.aws_profile_changed)

    def _build_profile_settings_ui(self, group, layout):
        layout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)

        self.job_history_dir_edit = DirectoryPickerWidget(
            initial_directory="",
            directory_label="Job history directory",
            parent=group,
            collapse_user_dir=True,
        )
        job_history_dir_label = self.labels["settings.job_history_dir"] = QLabel(
            "Job history directory"
        )
        layout.addRow(job_history_dir_label, self.job_history_dir_edit)
        self.job_history_dir_edit.path_changed.connect(self.job_history_dir_changed)

        self.default_farm_box = DeadlineFarmListComboBox(parent=group)
        default_farm_box_label = self.labels["defaults.farm_id"] = QLabel("Default farm")
        self.default_farm_box.box.currentIndexChanged.connect(self.default_farm_changed)
        self.default_farm_box.background_exception.connect(self.handle_background_exception)
        layout.addRow(default_farm_box_label, self.default_farm_box)

    def _build_farm_settings_ui(self, group, layout):
        layout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)

        self.default_queue_box = DeadlineQueueListComboBox(parent=group)
        default_queue_box_label = self.labels["defaults.queue_id"] = QLabel("Default queue")
        self.default_queue_box.box.currentIndexChanged.connect(self.default_queue_changed)
        self.default_queue_box.background_exception.connect(self.handle_background_exception)
        layout.addRow(default_queue_box_label, self.default_queue_box)

        self.default_storage_profile_box = DeadlineStorageProfileNameListComboBox(parent=group)
        default_storage_profile_box_label = self.labels["settings.storage_profile_id"] = QLabel(
            "Default storage profile"
        )
        self.default_storage_profile_box.box.currentIndexChanged.connect(
            self.default_storage_profile_name_changed
        )
        self.default_storage_profile_box.background_exception.connect(
            self.handle_background_exception
        )
        layout.addRow(default_storage_profile_box_label, self.default_storage_profile_box)

        item_name_copied = JobAttachmentsFileSystem.COPIED.value
        item_name_virtual = JobAttachmentsFileSystem.VIRTUAL.value
        job_attachments_file_system_tooltip = (
            "This setting determines how job attachments are loaded on the worker instance. "
            f"'{item_name_copied}' may be faster if every task needs all attachments, while "
            f"'{item_name_virtual}' may perform better if tasks only require a subset of attachments."
        )
        values_with_tooltips = {
            item_name_copied: "When selected, the worker downloads all job attachments to disk before rendering begins.",
            item_name_virtual: "When selected, the worker downloads attachments only when needed by each task.",
        }
        self.job_attachments_file_system_box = self._init_combobox_setting_with_tooltips(
            group=group,
            layout=layout,
            setting_name="defaults.job_attachments_file_system",
            label_text="Job attachments filesystem options",
            label_tooltip=job_attachments_file_system_tooltip,
            values_with_tooltips=values_with_tooltips,
        )

    def _build_general_settings_ui(self, group, layout):
        layout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)

        self.auto_accept = self._init_checkbox_setting(
            group, layout, "settings.auto_accept", "Auto accept prompt defaults"
        )
        self.telemetry_opt_out = self._init_checkbox_setting(
            group, layout, "telemetry.opt_out", "Telemetry opt out"
        )

        self._conflict_resolution_options = [option.name for option in FileConflictResolution]
        self.conflict_resolution_box = self._init_combobox_setting(
            group,
            layout,
            "settings.conflict_resolution",
            "Conflict resolution option",
            self._conflict_resolution_options,
        )

        self._log_levels = ["ERROR", "WARNING", "INFO", "DEBUG"]
        self.log_level_box = self._init_combobox_setting(
            group,
            layout,
            "settings.log_level",
            "Current logging level",
            self._log_levels,
        )

        # Known asset paths section
        known_paths_label = QLabel("Known asset paths")
        known_paths_label.setToolTip(
            "Paths that should not generate warnings when outside storage profile locations"
        )
        self.labels["settings.known_asset_paths"] = known_paths_label

        known_paths_widget = QWidget(parent=group)
        known_paths_widget.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        known_paths_layout = QVBoxLayout(known_paths_widget)
        known_paths_layout.setContentsMargins(0, 0, 0, 0)

        # Add buttons and status label
        button_layout = QHBoxLayout()
        self.add_known_path_button = QPushButton("Add...")
        self.add_known_path_button.clicked.connect(self._on_add_known_path)
        self.edit_known_path_button = QPushButton("Edit...")
        self.edit_known_path_button.clicked.connect(self._on_edit_known_path)
        self.remove_known_path_button = QPushButton("Remove Selected")
        self.remove_known_path_button.clicked.connect(self._on_remove_known_path)
        self.known_paths_status = QLabel()
        button_layout.addWidget(self.add_known_path_button)
        button_layout.addWidget(self.edit_known_path_button)
        button_layout.addWidget(self.remove_known_path_button)
        button_layout.addWidget(self.known_paths_status)
        known_paths_layout.addLayout(button_layout)

        # Add path list
        self.known_paths_list = QListWidget(parent=known_paths_widget)
        self.known_paths_list.setAlternatingRowColors(True)
        known_paths_layout.addWidget(self.known_paths_list)

        layout.addRow(known_paths_label, known_paths_widget)

        # Add refresh callback for known paths
        def refresh_known_paths():
            with block_signals(self.known_paths_list):
                self.known_paths_list.clear()
                paths_str = config_file.get_setting(
                    "settings.known_asset_paths", config=self.config
                )
                if paths_str:
                    paths = paths_str.split(os.pathsep)
                    self.known_paths_list.addItems(paths)

            # Update status label
            count = self.known_paths_list.count()
            self.known_paths_status.setText(f"{count} paths")

            self._update_known_paths_buttons()

        self._refresh_callbacks.append(refresh_known_paths)
        self.known_paths_list.itemSelectionChanged.connect(self._update_known_paths_buttons)

    def _init_checkbox_setting(
        self, group: QWidget, layout: QFormLayout, setting_name: str, label_text: str
    ) -> QCheckBox:
        """
        Creates a checkbox setting and adds it to the specified group and layout. This function also connects state
        change logic, refresh logic, and logic to save the setting in the config file.

        Args:
            group (QWidget): The parent of the checkbox
            layout (QFormLayout): The layout to add a row to for the checkbox
            setting_name (str): The setting name as provided to the config. E.g. "settings.foo_bar"
            label_text (str): The displayed description. E.g. "Foo Bar"

        Returns:
            QCheckBox: The created checkbox.
        """
        checkbox = QCheckBox(parent=group)
        label = self.labels[setting_name] = QLabel(label_text)
        layout.addRow(label, checkbox)

        def refresh_checkbox():
            """Function that refreshes the state of the checkbox based on the setting name"""
            with block_signals(checkbox):
                try:
                    state = str2bool(config_file.get_setting(setting_name, config=self.config))
                except ValueError as e:
                    logger.warning(f"{e} for '{setting_name}'")
                    state = False
                checkbox.setChecked(state)

        self._refresh_callbacks.append(refresh_checkbox)

        def checkbox_changed(new_state: int):
            """Callback for if the state of a given checkbox has changed"""
            value = str(checkbox.isChecked())
            self.changes[setting_name] = value
            self.refresh()

        checkbox.stateChanged.connect(checkbox_changed)

        return checkbox

    def _init_combobox_setting(
        self,
        group: QWidget,
        layout: QFormLayout,
        setting_name: str,
        label_text: str,
        values: List[str],
    ):
        """
        Creates a combobox setting and adds it to the specified group and layout. This function also connects state
        change logic, refresh logic, and logic to save the setting in the config file.

        Args:
            group (QWidget): The parent of the combobox
            layout (QFormLayout): The layout to add a row to for the combobox
            setting_name (str): The setting name as provided to the config. E.g. "settings.foo_bar"
            label_text (str): The displayed description. E.g. "Foo Bar"
            values (List[str]): The list of values that can be selected. E.g. ["Option A", "Option B"]

        Returns:
            QComboBox: The created combobox.
        """
        label = QLabel(label_text)
        combo_box = QComboBox(parent=group)
        layout.addRow(label, combo_box)
        combo_box.addItems(values)

        def refresh_combo_box():
            """Function that refreshes the state of the combo box based on the setting name"""
            with block_signals(combo_box):
                value = config_file.get_setting(setting_name, config=self.config)
                if value not in values:
                    default = get_setting_default(setting_name, config=self.config)
                    logger.warning(f"'{value}' is not one of {values}. Defaulting to '{default}'.")
                    value = default
                combo_box.setCurrentText(value)

        def combo_box_changed(new_value):
            """Callback for if the state of a given checkbox has changed"""
            self.changes[setting_name] = new_value
            self.refresh()

        combo_box.currentTextChanged.connect(combo_box_changed)

        self._refresh_callbacks.append(refresh_combo_box)

    def _init_combobox_setting_with_tooltips(
        self,
        group: QWidget,
        layout: QFormLayout,
        setting_name: str,
        label_text: str,
        label_tooltip: str,
        values_with_tooltips: Dict[str, str],
    ):
        """
        Creates and adds a combo box setting to the given group and layout, similar to `_init_combobox_setting`
        method. This method differentiates itself by adding tooltips for label and combo box items. Also,
        appends an (PySide6's built-in) Information icon at the label end to indicate tooltip availability.

        Args:
            group (QWidget): The parent of the combobox
            layout (QFormLayout): The layout to add a row to for the combobox
            setting_name (str): The setting name as provided to the config. E.g. "settings.foo_bar"
            label_text (str): The displayed description. E.g. "Foo Bar"
            label_tooltip (str): The tooltip for the label.
            values_with_tooltips (Dict[str, str]): The list of values that can be selected, along with their
                tooltips. E.g. {"Option A": "If A is selected, ...", "Option B": "If B is selected, ..."}
        """
        label = QLabel(label_text)
        icon_label = QLabel()
        icon = QApplication.style().standardIcon(QStyle.SP_MessageBoxInformation)
        icon_label.setPixmap(icon.pixmap(16, 16))
        icon_label.setToolTip(label_tooltip)

        combo_box = QComboBox(parent=group)

        row_layout = QHBoxLayout()
        row_layout.addWidget(label)
        row_layout.addWidget(icon_label)
        row_layout.addWidget(combo_box)
        row_layout.setStretchFactor(combo_box, 1)
        row_layout.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)

        layout.addRow(row_layout)

        for index, (value, tooltip) in enumerate(values_with_tooltips.items()):
            combo_box.addItem(value)
            combo_box.setItemData(index, tooltip, Qt.ToolTipRole)

        def refresh_combo_box():
            """Function that refreshes the state of the combo box based on the setting name"""
            with block_signals(combo_box):
                value = config_file.get_setting(setting_name, config=self.config)
                values = list(values_with_tooltips.keys())
                if value not in values:
                    default = get_setting_default(setting_name, config=self.config)
                    logger.warning(f"'{value}' is not one of {values}. Defaulting to '{default}'.")
                    value = default
                combo_box.setCurrentText(value)
                combo_box.setToolTip(values_with_tooltips[value])

        def combo_box_changed(new_value):
            """Callback for if the state of a given checkbox has changed"""
            self.changes[setting_name] = new_value
            self.refresh()

        combo_box.currentTextChanged.connect(combo_box_changed)

        self._refresh_callbacks.append(refresh_combo_box)

    def handle_background_exception(self, title, e):
        QMessageBox.warning(self, title, f"Encountered an error:\n{e}")  # type: ignore[call-arg]

    def _fill_aws_profiles_box(self):
        # Use boto3 directly with no profile, so we don't get an error
        # if the configured profile does not exist.
        try:
            session = boto3.Session()
            aws_profile_names = [
                "(default)",
                *(
                    name
                    for name in session._session.full_config["profiles"].keys()
                    if name != "default"
                ),
            ]
        except ProfileNotFound:
            logger.exception("Failed to create boto3.Session for AWS profile list")
            aws_profile_names = [f"{NOT_VALID_MARKER} <failed to retrieve AWS profile names>"]

        self.aws_profile_names = aws_profile_names
        with block_signals(self.aws_profiles_box):
            self.aws_profiles_box.addItems(list(self.aws_profile_names))

    def refresh_lists(self):
        if api.check_deadline_api_available():
            self.default_farm_box.refresh_list()
            self.default_queue_box.refresh_list()
            self.default_storage_profile_box.refresh_list()

    def refresh(self):
        """
        Refreshes all the configuration UI elements from the current config.
        """
        # Make self.config be a deep copy of the config, with changes applied
        self.config = ConfigParser()
        self.config.read_dict(config_file.read_config())
        for setting_name, value in self.changes.items():
            config_file.set_setting(setting_name, value, self.config)
        self.default_farm_box.set_config(self.config)
        self.default_queue_box.set_config(self.config)
        self.default_storage_profile_box.set_config(self.config)

        with block_signals(self.aws_profiles_box):
            aws_profile_name = config_file.get_setting(
                "defaults.aws_profile_name", config=self.config
            )
            # Change the values representing the default to the UI value representing the default
            if aws_profile_name in ("(default)", "default", ""):
                aws_profile_name = "(default)"
            elif aws_profile_name not in self.aws_profile_names:
                aws_profile_name = f"{NOT_VALID_MARKER} {aws_profile_name}"
            index = self.aws_profiles_box.findText(aws_profile_name, Qt.MatchFixedString)
            if index >= 0:
                self.aws_profiles_box.setCurrentIndex(index)
            else:
                self.aws_profiles_box.insertItem(0, aws_profile_name)
                self.aws_profiles_box.setCurrentIndex(0)

        with block_signals(self.job_history_dir_edit):
            job_history_dir = config_file.get_setting(
                "settings.job_history_dir", config=self.config
            )
            self.job_history_dir_edit.setText(job_history_dir)

        self.default_farm_box.refresh_selected_id()

        for refresh_callback in self._refresh_callbacks:
            refresh_callback()

        self.default_queue_box.refresh_selected_id()
        self.default_storage_profile_box.refresh_selected_id()

        # Put an orange box around the labels for any settings that are changed
        for setting_name, label in self.labels.items():
            if setting_name in self.changes:
                label.setStyleSheet("border: 1px solid orange;")
            else:
                label.setStyleSheet("")

        self.refreshed.emit()

    def apply(self) -> bool:
        """
        Apply all the settings that the user has changed into the config file.

        Returns True if the settings were applied, False otherwise.
        """

        # We need to retrieve here as changing Queue's won't update.
        self.changes["settings.storage_profile_id"] = (
            self.default_storage_profile_box.box.currentData()
        )

        for setting_name, value in self.changes.items():
            if value.startswith(NOT_VALID_MARKER):
                QMessageBox.warning(  # type: ignore[call-arg]
                    self,
                    "Apply changes",
                    f"Cannot apply changes, {value} is not valid for setting {setting_name}",
                )
                return False

        self.config = config_file.read_config()

        for setting_name, value in self.changes.items():
            config_file.set_setting(setting_name, value, self.config)
        root.setLevel(config_file.get_setting("settings.log_level"))
        api.get_deadline_cloud_library_telemetry_client().set_opt_out(config=self.config)

        # Only update self.changes_were_applied if false. We don't want to invalidate that a change has
        # occurred if the user repeatedly hits "Apply" or hits "Apply" and then "Save".
        if not self.changes_were_applied:
            self.changes_were_applied = len(self.changes) > 0

        self.changes.clear()

        config_file.write_config(self.config)

        # Refresh the GUI (writing the config file should cause this, but do this redundantly to make sure)
        self.refresh()

        return True

    def aws_profile_changed(self, value):
        self.changes["defaults.aws_profile_name"] = value
        self.default_farm_box.clear_list()
        self.default_queue_box.clear_list()
        self.default_storage_profile_box.clear_list()
        self.refresh()

    def job_history_dir_changed(self):
        job_history_dir = self.job_history_dir_edit.text()
        # Only apply the change if the text was actually edited
        if job_history_dir != config_file.get_setting(
            "settings.job_history_dir", config=self.config
        ):
            self.changes["settings.job_history_dir"] = job_history_dir
        self.refresh()

    def default_farm_changed(self, index):
        self.changes["defaults.farm_id"] = self.default_farm_box.box.itemData(index)
        self.refresh()
        self.default_queue_box.refresh_list()
        self.default_storage_profile_box.refresh_list()

    def default_queue_changed(self, index):
        self.changes["defaults.queue_id"] = self.default_queue_box.box.itemData(index)
        self.refresh()
        self.default_storage_profile_box.refresh_list()

    def default_storage_profile_name_changed(self, index):
        self.changes["settings.storage_profile_id"] = self.default_storage_profile_box.box.itemData(
            index
        )
        self.refresh()

    def _on_add_known_path(self):
        """Handle adding a new known path"""
        path = QFileDialog.getExistingDirectory(
            self, "Select a directory to add to known asset paths", os.path.expanduser("~")
        )
        if path:
            # Normalize path
            path = os.path.normpath(path)

            # Get current paths
            current_paths = []
            for i in range(self.known_paths_list.count()):
                current_paths.append(self.known_paths_list.item(i).text())

            # Only add if not already in list
            if path not in current_paths:
                self.known_paths_list.addItem(path)

                # Update config
                current_paths.append(path)
                self.changes["settings.known_asset_paths"] = os.pathsep.join(current_paths)
                self.refresh()

    def _on_remove_known_path(self):
        """Handle removing the selected known path"""
        current_row = self.known_paths_list.currentRow()
        if current_row >= 0:
            self.known_paths_list.takeItem(current_row)

            # Update config
            current_paths = []
            for i in range(self.known_paths_list.count()):
                current_paths.append(self.known_paths_list.item(i).text())
            self.changes["settings.known_asset_paths"] = os.pathsep.join(current_paths)
            self.refresh()

    def _update_known_paths_buttons(self):
        """Enable/disable edit and remove buttons based on selection"""
        has_selection = self.known_paths_list.currentRow() >= 0
        self.edit_known_path_button.setEnabled(has_selection)
        self.remove_known_path_button.setEnabled(has_selection)

    def _on_edit_known_path(self):
        """Handle editing the selected known path"""
        current_row = self.known_paths_list.currentRow()
        if current_row >= 0:
            current_path = self.known_paths_list.item(current_row).text()
            path = QFileDialog.getExistingDirectory(
                self,
                "Select a directory to replace the selected path",
                current_path if os.path.exists(current_path) else os.path.expanduser("~"),
            )
            if path:
                # Normalize path
                path = os.path.normpath(path)

                # Only update if path changed
                if path != current_path:
                    # Get current paths
                    current_paths = []
                    for i in range(self.known_paths_list.count()):
                        if i != current_row:  # Skip the path being edited
                            current_paths.append(self.known_paths_list.item(i).text())

                    # Only add if not already in list
                    if path not in current_paths:
                        self.known_paths_list.item(current_row).setText(path)
                        current_paths.insert(current_row, path)

                    self.changes["settings.known_asset_paths"] = os.pathsep.join(current_paths)
                    self.refresh()


class _DeadlineResourceListComboBox(QWidget):
    """
    A ComboBox for selecting an AWS Deadline Cloud Id, with a refresh button.

    The caller should connect the `background_exception` signal, e.g.
    to show a message box, and should call `set_config` whenever there is
    a change to the AWS Deadline Cloud config object.

    Args:
        resource_name (str): The resource name for the list, like "Farm",
                "Queue", "Fleet".
    """

    # Emitted when the background refresh thread catches an exception,
    # provides (operation_name, BaseException)
    background_exception = Signal(str, BaseException)

    # Emitted when an async refresh_farms_list thread completes,
    # provides (refresh_id, [(farm_id, farm_name), ...])
    _list_update = Signal(int, list)

    def __init__(self, resource_name, setting_name, parent=None):
        super().__init__(parent)

        self.__refresh_thread = None
        self.__refresh_id = 0
        self.canceled = CancelationFlag()
        self.destroyed.connect(self.canceled.set_canceled)

        self.resource_name = resource_name
        self.setting_name = setting_name

        self._build_ui()

    def _build_ui(self):
        self.box = QComboBox(parent=self)
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.box, stretch=1)

        self.refresh_button = QPushButton("")
        layout.addWidget(self.refresh_button)
        self.refresh_button.setIcon(QApplication.style().standardIcon(QStyle.SP_BrowserReload))
        self.refresh_button.setFixedSize(QSize(22, 22))  # Make the button square
        self.refresh_button.clicked.connect(self.refresh_list)
        self._list_update.connect(self.handle_list_update)
        self.background_exception.connect(self.handle_background_exception)

    def handle_background_exception(self, e):
        with block_signals(self.box):
            self.box.clear()
        self.refresh_selected_id()

    def count(self) -> int:
        """Returns the number of items in the combobox"""
        return self.box.count()

    def set_config(self, config: ConfigParser) -> None:
        """Updates the AWS Deadline Cloud config object the control uses."""
        self.config = config

    def clear_list(self):
        """
        Fully clears the list. The caller needs to call either
        `refresh_list` or `refresh_selected_id` at a later point
        to finish it.
        """
        with block_signals(self.box):
            self.box.clear()

    def refresh_list(self):
        """
        Starts a background thread to refresh the resource list.
        """
        config = self.config
        selected_id = config_file.get_setting(self.setting_name, config=config)
        # Reset to a list of just the currently configured id during refresh
        with block_signals(self.box):
            self.box.clear()
            self.box.addItem("<refreshing>", userData=selected_id)

        self.__refresh_id += 1
        self.__refresh_thread = threading.Thread(
            target=self._refresh_thread_function,
            name=f"AWS Deadline Cloud refresh {self.resource_name} thread",
            args=(self.__refresh_id, config),
        )
        self.__refresh_thread.start()

    def handle_list_update(self, refresh_id, items_list):
        # Apply the refresh if it's still for the latest call
        if refresh_id == self.__refresh_id:
            with block_signals(self.box):
                self.box.clear()
                for name, id in items_list:
                    self.box.addItem(name, userData=id)

                self.refresh_selected_id()

    def refresh_selected_id(self):
        """Refreshes the selected id from the config object"""
        selected_id = config_file.get_setting(self.setting_name, config=self.config)
        # Restore the selected Id, inserting a new item if
        # it doesn't exist in the list.
        with block_signals(self.box):
            index = self.box.findData(selected_id)
            if index >= 0:
                self.box.setCurrentIndex(index)
            else:
                # Some cases allow to select "nothing" and insert an item to indicate such
                index = self.box.findText("<none selected>")
                if index >= 0:
                    self.box.setCurrentIndex(index)
                else:
                    self.box.insertItem(0, "<none selected>", userData="")
                    self.box.setCurrentIndex(0)

    def _refresh_thread_function(self, refresh_id: int, config: Optional[ConfigParser] = None):
        """
        This function gets started in a background thread to refresh the list.
        """
        try:
            resources = self.list_resources(config=config)
            if not self.canceled:
                self._list_update.emit(refresh_id, resources)
        except BaseException as e:
            if not self.canceled and refresh_id == self.__refresh_id:
                self.background_exception.emit(f"Refresh {self.resource_name}s list", e)


class DeadlineFarmListComboBox(_DeadlineResourceListComboBox):
    def __init__(self, parent=None):
        super().__init__(resource_name="Farm", setting_name="defaults.farm_id", parent=parent)

    def list_resources(self, config: Optional[ConfigParser]):
        response = api.list_farms(config=config)
        return sorted(
            [(item["displayName"], item["farmId"]) for item in response["farms"]],
            key=lambda item: (item[0].casefold(), item[1]),
        )


class DeadlineQueueListComboBox(_DeadlineResourceListComboBox):
    def __init__(self, parent=None):
        super().__init__(resource_name="Queue", setting_name="defaults.queue_id", parent=parent)

    def list_resources(self, config: Optional[ConfigParser]):
        default_farm_id = config_file.get_setting("defaults.farm_id", config=config)
        if default_farm_id:
            response = api.list_queues(config=config, farmId=default_farm_id)
            return sorted(
                [(item["displayName"], item["queueId"]) for item in response["queues"]],
                key=lambda item: (item[0].casefold(), item[1]),
            )
        else:
            return []


class DeadlineStorageProfileNameListComboBox(_DeadlineResourceListComboBox):
    WINDOWS_OS = "windows"
    MAC_OS = "macos"
    LINUX_OS = "linux"

    def __init__(self, parent=None):
        super().__init__(
            resource_name="Storage profile",
            setting_name="settings.storage_profile_id",
            parent=parent,
        )

    def list_resources(self, config: Optional[ConfigParser]):
        default_farm_id = config_file.get_setting("defaults.farm_id", config=config)
        default_queue_id = config_file.get_setting("defaults.queue_id", config=config)
        if default_farm_id and default_queue_id:
            response = api.list_storage_profiles_for_queue(
                config=config, farmId=default_farm_id, queueId=default_queue_id
            )
            storage_profiles = response.get("storageProfiles", [])
            # add a "<none selected>" option since its possible to select nothing for this type
            # of resource
            storage_profiles.append(
                {
                    "storageProfileId": "",
                    "displayName": "<none selected>",
                    "osFamily": self._get_current_os(),
                }
            )
            return sorted(
                [
                    (item["displayName"], item["storageProfileId"])
                    for item in storage_profiles
                    if self._get_current_os() == item["osFamily"].lower()
                ],
                key=lambda item: (item[0].casefold(), item[1]),
            )
        else:
            return []

    def _get_current_os(self) -> str:
        """
        Get a string specifying what the OS is, following the format the Deadline storage profile API expects.
        """
        if sys.platform.startswith("linux"):
            return self.LINUX_OS

        if sys.platform.startswith("darwin"):
            return self.MAC_OS

        if sys.platform.startswith("win"):
            return self.WINDOWS_OS

        return "Unknown"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Provides functionality for boto3 Sessions, Clients, and properties
of the Deadline-configured IAM credentials.
"""

from __future__ import annotations

import logging
from configparser import ConfigParser
from contextlib import contextmanager
from enum import Enum
from functools import lru_cache
from typing import Optional, Tuple

import boto3  # type: ignore[import]
import botocore
import botocore.config
from botocore.client import BaseClient  # type: ignore[import]
from botocore.credentials import CredentialProvider, RefreshableCredentials
from botocore.exceptions import (  # type: ignore[import]
    ClientError,
    ProfileNotFound,
)
from botocore.session import get_session as get_botocore_session

from .. import version
from ..config import get_setting
from ..exceptions import DeadlineOperationError
from ...job_attachments._aws.aws_clients import get_s3_client


class AwsCredentialsSource(Enum):
    NOT_VALID = 0
    HOST_PROVIDED = 2
    DEADLINE_CLOUD_MONITOR_LOGIN = 3


class AwsAuthenticationStatus(Enum):
    CONFIGURATION_ERROR = 1
    AUTHENTICATED = 2
    NEEDS_LOGIN = 3


# Place for stashing context to be attached to boto clients.
session_context: dict[str, Optional[str]] = {
    "submitter-name": None,
    "cli-command-name": None,
}


def get_boto3_session(
    force_refresh: bool = False, config: Optional[ConfigParser] = None
) -> boto3.Session:
    """
    Gets a boto3 session for the configured AWS Deadline Cloud aws profile. This may
    either use a named profile or the default credentials provider chain.

    This implementation caches the session object for use across the CLI code,
    so that we can use the following code pattern without repeated calls to
    an external credentials provider process, for example.

    Args:
        force_refresh (bool, optional): If set to True, forces a cache refresh.
        config (ConfigParser, optional): If provided, the AWS Deadline Cloud config to use.
    """

    profile_name: Optional[str] = get_setting("defaults.aws_profile_name", config)

    # If the default AWS profile name is either not set, or set to "default",
    # use the default credentials provider chain instead of a named profile.
    if profile_name in ("(default)", "default", ""):
        profile_name = None

    if force_refresh:
        invalidate_boto3_session_cache()

    return _get_boto3_session_for_profile(profile_name)


@lru_cache
def _get_boto3_session_for_profile(profile_name: str):
    session = boto3.Session(profile_name=profile_name)

    # By default, DCM returns creds that expire after 15 minutes, and boto3's RefreshableCredentials
    # class refreshes creds that are within 15 minutes of expiring, so credentials would never be reused.
    # Also DCM credentials currently take several seconds to refresh. Lower the refresh timeouts so creds
    # are reused between API calls to save time.
    # See https://github.com/boto/botocore/blob/develop/botocore/credentials.py#L342-L362

    try:
        credentials = session.get_credentials()
        if (
            isinstance(credentials, RefreshableCredentials)
            and credentials.method == "custom-process"
        ):
            credentials._advisory_refresh_timeout = 5 * 60  # 5 minutes
            credentials._mandatory_refresh_timeout = 2.5 * 60  # 2.5 minutes
    except:  # noqa: E722
        # Attempt to patch the timeouts but ignore any errors. These patched proeprties are internal and could change
        # without notice. Creds are functional without patching timeouts.
        pass

    return session


def invalidate_boto3_session_cache() -> None:
    _get_boto3_session_for_profile.cache_clear()
    _get_queue_user_boto3_session.cache_clear()


def get_default_client_config(**kwargs) -> botocore.config.Config:
    """
    Gets the default botocore Config object to use with `boto3 clients`.
    This method adds user agent version and submitter context into botocore calls.
    Additional arguments are forwarded to the Config constructor.
    """
    user_agent_extra = f"app/deadline-client#{version}"
    if session_context.get("submitter-name"):
        user_agent_extra += f" submitter/{session_context['submitter-name']}"
    if session_context.get("cli-command-name"):
        user_agent_extra += f" cli-command/{session_context['cli-command-name']}"
    client_config = botocore.config.Config(user_agent_extra=user_agent_extra, **kwargs)
    return client_config


@lru_cache
def get_session_client(session: boto3.Session, service_name: str):
    """
    Create and cache a boto3 client for the given session and service name.

    This function is decorated with @lru_cache to ensure that repeated calls
    with the same session and service name return the cached client to avoid
    repeating initialization where possible.

    Args:
        session: The boto3 Session to use for creating the client
        service_name: The name of the AWS service (e.g., 's3', 'sts', 'ec2')

    Returns:
        A boto3 client for the specified service
    """
    return session.client(service_name, config=get_default_client_config())


def get_boto3_client(service_name: str, config: Optional[ConfigParser] = None) -> BaseClient:
    """
    Gets a client from the boto3 session returned by `get_boto3_session`.
    If the client requested is `deadline`, it uses the AWS_ENDPOINT_URL_DEADLINE
    deadline endpoint url.

    Args:
        service_name (str): The AWS service to get the client for, e.g. "deadline".
        config (ConfigParser, optional): If provided, the AWS Deadline Cloud config to use.
    """

    session = get_boto3_session(config=config)
    return get_session_client(session=session, service_name=service_name)


def get_credentials_source(config: Optional[ConfigParser] = None) -> AwsCredentialsSource:
    """
    Returns DEADLINE_CLOUD_MONITOR_LOGIN if Deadline Cloud monitor wrote the credentials, HOST_PROVIDED otherwise.

    Args:
        config (ConfigParser, optional): The AWS Deadline Cloud configuration
                object to use instead of the config file.
    """
    try:
        session = get_boto3_session(config=config)
        profile_config = session._session.get_scoped_config()
    except ProfileNotFound:
        return AwsCredentialsSource.NOT_VALID
    if "monitor_id" in profile_config:
        # Deadline Cloud monitor Desktop adds the "monitor_id" key
        return AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN
    else:
        return AwsCredentialsSource.HOST_PROVIDED


def get_user_and_identity_store_id(
    config: Optional[ConfigParser] = None,
) -> tuple[Optional[str], Optional[str]]:
    """
    If logged in with Deadline Cloud monitor Desktop, returns a tuple
    (user_id, identity_store_id), otherwise returns None.
    """
    session = get_boto3_session(config=config)
    profile_config = session._session.get_scoped_config()

    if "monitor_id" in profile_config:
        return (profile_config["user_id"], profile_config["identity_store_id"])
    else:
        return None, None


def get_monitor_id(
    config: Optional[ConfigParser] = None,
) -> Optional[str]:
    """
    If logged in with Deadline Cloud Monitor to a Deadline Monitor, returns Monitor Id, otherwise returns None.
    """
    session = get_boto3_session(config=config)
    profile_config = session._session.get_scoped_config()

    return profile_config.get("monitor_id", None)


def get_queue_user_boto3_session(
    deadline: BaseClient,
    config: Optional[ConfigParser] = None,
    farm_id: Optional[str] = None,
    queue_id: Optional[str] = None,
    queue_display_name: Optional[str] = None,
    force_refresh: bool = False,
) -> boto3.Session:
    """
    Calls the AssumeQueueRoleForUser API to obtain the role configured in a Queue,
    and then creates and returns a boto3 session with those credentials.

    Args:
        deadline (BaseClient): A Deadline client.
        config (ConfigParser, optional): If provided, the AWS Deadline Cloud config to use.
        farm_id (str, optional): The ID of the farm to use.
        queue_id (str, optional): The ID of the queue to use.
        queue_display_name (str, optional): The display name of the queue.
        force_refresh (bool, optional): If True, forces a cache refresh.
    """

    base_session = get_boto3_session(config=config, force_refresh=force_refresh)

    if farm_id is None:
        farm_id = get_setting("defaults.farm_id")
    if queue_id is None:
        queue_id = get_setting("defaults.queue_id")

    return _get_queue_user_boto3_session(
        deadline, base_session, farm_id, queue_id, queue_display_name
    )


@lru_cache
def _get_queue_user_boto3_session(
    deadline: BaseClient,
    base_session: boto3.Session,
    farm_id: str,
    queue_id: str,
    queue_display_name: Optional[str] = None,
):
    queue_credential_provider = QueueUserCredentialProvider(
        deadline,
        farm_id,
        queue_id,
        queue_display_name,
    )

    botocore_session = get_botocore_session()
    credential_provider = botocore_session.get_component("credential_provider")
    credential_provider.insert_before("env", queue_credential_provider)
    aws_profile_name: Optional[str] = None
    if base_session.profile_name != "default":
        aws_profile_name = base_session.profile_name

    return boto3.Session(
        botocore_session=botocore_session,
        profile_name=aws_profile_name,
        region_name=base_session.region_name,
    )


@contextmanager
def _modified_logging_level(logger, level):
    old_level = logger.getEffectiveLevel()
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(old_level)


def check_authentication_status(config: Optional[ConfigParser] = None) -> AwsAuthenticationStatus:
    """
    Checks the status of the provided session, by
    calling the sts::GetCallerIdentity API.

    Args:
        config (ConfigParser, optional): The AWS Deadline Cloud configuration
                object to use instead of the config file.

    Returns AwsAuthenticationStatus enum value:
      - CONFIGURATION_ERROR if there is an unexpected error accessing credentials
      - AUTHENTICATED if they are fine
      - NEEDS_LOGIN if a Deadline Cloud monitor login is required.
    """

    with _modified_logging_level(logging.getLogger("botocore.credentials"), logging.ERROR):
        try:
            get_boto3_session(config=config).client("sts").get_caller_identity()
            return AwsAuthenticationStatus.AUTHENTICATED
        except Exception:
            # We assume that the presence of a Deadline Cloud monitor profile
            # means we will know everything necessary to start it and login.

            if get_credentials_source(config) == AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN:
                return AwsAuthenticationStatus.NEEDS_LOGIN
            return AwsAuthenticationStatus.CONFIGURATION_ERROR


def precache_clients(
    deadline: BaseClient = None,
    config: Optional[ConfigParser] = None,
    farm_id: Optional[str] = None,
    queue_id: Optional[str] = None,
    queue_display_name: Optional[str] = None,
) -> Tuple[BaseClient, BaseClient]:
    """
    Initialize an S3 client (and optionally a Deadline client) with queue user credentials
    to pre-warm the client cache.

    This function creates an S3 client using queue user credentials, which triggers
    the expensive service discovery process once. Subsequent client creations using
    the same session object should then use the cached client, improving performance.
    This function is designed to be called in a background thread at application startup.

    Args:
        deadline: An existing deadline client. If None, one will be created.
        config: Optional configuration parser. If None, the default configuration will be used.
        farm_id: The farm ID. If None, it will be retrieved from settings.
        queue_id: The queue ID. If None, it will be retrieved from settings.
        queue_display_name: The queue display name. If None, it will be retrieved from the queue.

    Returns:
        Created (or current) s3 client for the given queue_role_session

    Example:
        # Fire and forget initialization in a background thread
        import threading
        threading.Thread(
            target=initialize_queue_user_s3_client,
            daemon=True,
            name="S3ClientInit"
        ).start()
    """
    if not deadline:
        deadline = get_boto3_client("deadline", config=config)
    if not queue_id:
        queue_id = get_setting("defaults.queue_id", config=config)
    if not farm_id:
        farm_id = get_setting("defaults.farm_id", config=config)

    if not queue_display_name:
        queue = deadline.get_queue(
            farmId=farm_id,
            queueId=queue_id,
        )
        queue_display_name = queue["displayName"]

    queue_role_session = get_queue_user_boto3_session(
        deadline=deadline,
        config=config,
        farm_id=farm_id,
        queue_id=queue_id,
        queue_display_name=queue_display_name,
    )
    # Initialize the S3 client to populate the cache
    return deadline, get_s3_client(queue_role_session)


class QueueUserCredentialProvider(CredentialProvider):
    """A custom botocore CredentialProvider for handling AssumeQueueRoleForUser API
    credentials. If the credentials expire, the provider will automatically refresh
    them using the _get_queue_credentials method.
    """

    # The following two constants are part of botocore's CredentialProvider interface

    # A short name to identify the provider within botocore.
    METHOD = "queue-credential-provider"
    # A name to identify the provider for use in cross-sdk features. The AWS SDKs
    # require that providers outside of botocore are prefixed with "custom"
    CANONICAL_NAME = "custom-queue-credential-provider"

    deadline: BaseClient
    farm_id: str
    queue_id: str
    queue_display_name_or_id: Optional[str]

    def __init__(
        self,
        deadline: BaseClient,
        farm_id: str,
        queue_id: str,
        queue_display_name: Optional[str] = None,
    ):
        self.deadline = deadline
        self.farm_id = farm_id
        self.queue_id = queue_id
        self.queue_display_name_or_id = queue_display_name or queue_id

    def load(self):
        credentials = self._get_queue_credentials()
        return RefreshableCredentials.create_from_metadata(
            metadata=credentials,
            refresh_using=self._get_queue_credentials,
            method=self.METHOD,
        )

    def _get_queue_credentials(self):
        """
        Fetches or refreshes the credentials using the AssumeQueueRoleForUser API
        for the specified Farm ID and Queue ID.
        """
        try:
            queue_credentials = self.deadline.assume_queue_role_for_user(
                farmId=self.farm_id, queueId=self.queue_id
            ).get("credentials", None)
        except ClientError as exc:
            code = exc.response.get("Error", {}).get("Code", None)
            if code == "ThrottlingException":
                raise DeadlineOperationError(
                    f"Throttled while attempting to assume Queue role for user on Queue '{self.queue_display_name_or_id}': {exc}\n"
                    "Please retry the operation later, or contact your administrator to increase the API's rate limit."
                ) from exc
            elif code == "InternalServerException":
                raise DeadlineOperationError(
                    f"An internal server error occurred while attempting to assume Queue role for user on "
                    f"Queue '{self.queue_display_name_or_id}': {exc}\n"
                ) from exc
            else:
                raise DeadlineOperationError(
                    f"Failed to assume Queue role for user on Queue '{self.queue_display_name_or_id}': {exc}\nPlease contact your "
                    "administrator to ensure a Queue role exists and that you have permissions to access this Queue."
                ) from exc
        if not queue_credentials:
            raise DeadlineOperationError(
                f"Failed to get credentials for '{self.queue_display_name_or_id}': Empty credentials received."
            )
        return {
            "access_key": queue_credentials["accessKeyId"],
            "secret_key": queue_credentials["secretAccessKey"],
            "token": queue_credentials["sessionToken"],
            "expiry_time": queue_credentials["expiration"].isoformat(),
        }



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Contains functions validating Asset Manifests version 2023_03_03."""

from __future__ import annotations

from typing import Any, Optional, Tuple

_REQUIRED_FIELDS_2023_03_03: list[str] = [
    "hashAlg",
    "paths",
    "manifestVersion",
    "totalSize",
]

_HASH_ALGS_2023_03_03: set[str] = {"xxh128"}

_PATH_REQUIRED_FIELDS_2023_03_03: list[str] = [
    "path",
    "hash",
    "size",
    "mtime",
]


def _get_missing_fields(obj: dict[str, Any], required: list[str]) -> list[str]:
    missing = []
    for field in required:
        if field not in obj:
            missing.append(field)
    return missing


def _validate_path_2023_03_03(path_object: dict[str, Any]) -> Tuple[bool, Optional[str]]:
    missing = _get_missing_fields(path_object, _PATH_REQUIRED_FIELDS_2023_03_03)
    if len(missing) > 0:
        return False, f"path is missing required field(s) {missing}"

    path = path_object["path"]
    if not isinstance(path, str):
        return False, "path must be a string"

    hash = path_object["hash"]
    if not isinstance(hash, str):
        return False, "hash must be a string"

    size = path_object["size"]
    if not isinstance(size, int):
        return False, "size must be an integer"

    mtime = path_object["mtime"]
    if not isinstance(mtime, int):
        return False, "mtime must be an integer"

    return True, None


def validate_manifest_2023_03_03(manifest: dict[str, Any]) -> Tuple[bool, Optional[str]]:
    missing = _get_missing_fields(manifest, _REQUIRED_FIELDS_2023_03_03)
    if len(missing) > 0:
        return False, f"manifest is missing required field(s) {missing}"

    manifest_version = manifest["manifestVersion"]
    if not isinstance(manifest_version, str) or manifest_version != "2023-03-03":
        return False, 'manifestVersion must be "2023-03-03"'

    hash_alg = manifest["hashAlg"]
    if not isinstance(hash_alg, str) or hash_alg not in _HASH_ALGS_2023_03_03:
        return False, f"hashAlg must be one of {_HASH_ALGS_2023_03_03}"

    total_size = manifest["totalSize"]
    if not isinstance(total_size, int):
        return False, "totalSize must be an integer"

    paths = manifest["paths"]
    if not isinstance(paths, list):
        return False, "paths must be a list"
    elif len(paths) < 1:
        return False, "paths must have a least one item"
    else:
        for path_object in paths:
            ok, message = _validate_path_2023_03_03(path_object)
            if not ok:
                return False, message

    return True, None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

__all__ = ["_incremental_output_download"]

from dataclasses import dataclass, asdict
from datetime import datetime, timedelta, timezone
import difflib
from typing import Optional
from configparser import ConfigParser
from typing import Any, Callable
import sys
import time
import concurrent.futures
import textwrap

from .. import api
import boto3
from botocore.client import BaseClient  # type: ignore[import]
from ..exceptions import DeadlineOperationError
from ..api._list_jobs_by_filter_expression import _list_jobs_by_filter_expression
from ..api._session import get_session_client
from ...common.path_utils import summarize_path_list, human_readable_file_size
from ...job_attachments._incremental_downloads.incremental_download_state import (
    IncrementalDownloadState,
    IncrementalDownloadJob,
    _datetimes_to_str,
)
from ...job_attachments._incremental_downloads._manifest_s3_downloads import (
    _add_output_manifests_from_s3,
    _download_all_manifests_with_absolute_paths,
    _merge_absolute_path_manifest_list,
    _download_manifest_paths,
)
from ...job_attachments._path_mapping import _generate_path_mapping_rules, _PathMappingRuleApplier
from ...job_attachments.asset_manifests import (
    BaseAssetManifest,
    BaseManifestPath,
)
from ...job_attachments.asset_manifests import (
    HashAlgorithm,
)
from ...job_attachments.models import (
    FileConflictResolution,
    PathFormat,
    StorageProfileOperatingSystemFamily,
)
from ...job_attachments.progress_tracker import (
    ProgressReportMetadata,
)
from ._common import _cli_object_repr, sigint_handler


SESSIONS_API_MAX_CONCURRENCY = 3


@dataclass
class IncrementalOutputDownloadLatencies:
    """Dataclass for tracking latencies of operations in this command"""

    _get_download_candidate_jobs: int = 0
    _categorize_jobs_in_checkpoint: int = 0
    _get_job_sessions: int = 0
    _update_checkpoint_jobs_list: int = 0
    _download_all_manifests_with_absolute_paths: int = 0
    download: Optional[int] = None
    path_mapping: Optional[int] = None


def _get_download_candidate_jobs(
    boto3_session: boto3.Session,
    farm_id: str,
    queue_id: str,
    starting_timestamp: datetime,
    print_function_callback: Callable[[str], None] = lambda msg: None,
) -> dict[str, dict[str, Any]]:
    """
    Uses deadline:SearchJobs queries to get a dict {job_id: job} of download candidates for the queue.
    This is a superset of all the jobs that have produced any output for download since
    the provided starting_timestamp.

    Args:
        boto3_session: The boto3.Session for accessing AWS.
        farm_id: The farm id for the operation.
        queue_id: The queue id for the operation.
        starting_timestamp: The point in time from which to look for new download outputs.
        print_function_callback: Callback for printing output to the terminal or log.

    Returns:
        A dictionary mapping job id to the job as returned by the deadline.search_jobs API.
    """
    print_function_callback("Retrieving updated data from Deadline Cloud...")
    start_time = datetime.now(tz=timezone.utc)

    # Construct the full set of jobs that may have new available downloads.
    # - Any active job (job with taskRunStatus in READY, ASSIGNED,
    #   STARTING, SCHEDULED, or RUNNING), that has at least one SUCCEEDED task.
    download_candidate_jobs = {
        job["jobId"]: job
        for job in _list_jobs_by_filter_expression(
            boto3_session,
            farm_id,
            queue_id,
            filter_expression={
                "filters": [
                    {
                        "stringFilter": {
                            "name": "TASK_RUN_STATUS",
                            "operator": "EQUAL",
                            "value": status_value,
                        },
                    }
                    # Maximum of 3 filters are permitted, so the 5 statuses are split
                    for status_value in ["READY", "ASSIGNED", "STARTING"]
                ],
                "operator": "OR",
            },
        )
    }
    download_candidate_jobs.update(
        {
            job["jobId"]: job
            for job in _list_jobs_by_filter_expression(
                boto3_session,
                farm_id,
                queue_id,
                filter_expression={
                    "filters": [
                        {
                            "stringFilter": {
                                "name": "TASK_RUN_STATUS",
                                "operator": "EQUAL",
                                "value": status_value,
                            },
                        }
                        for status_value in ["SCHEDULED", "RUNNING"]
                    ],
                    "operator": "OR",
                },
            )
        }
    )
    print(f"DEBUG: Got {len(download_candidate_jobs)} active jobs")
    download_candidate_jobs = {
        job_id: _datetimes_to_str(job)
        for job_id, job in download_candidate_jobs.items()
        if job["taskRunStatusCounts"]["SUCCEEDED"] > 0
    }
    print(
        f"DEBUG: Filtered down to {len(download_candidate_jobs)} active jobs based on SUCCEEDED task filter"
    )

    # - Any recently ended job (job went from active to terminal with a taskRunStatus
    #   in SUSPENDED, CANCELED, FAILED, SUCCEEDED, NOT_COMPATIBLE), that has at least
    #   one SUCCEEDED task. The endedAt timestamp field gets updated when that occurs.
    recently_ended_jobs = _list_jobs_by_filter_expression(
        boto3_session,
        farm_id,
        queue_id,
        filter_expression={
            "filters": [
                {
                    "dateTimeFilter": {
                        "name": "ENDED_AT",
                        "dateTime": starting_timestamp,
                        "operator": "GREATER_THAN_EQUAL_TO",
                    }
                }
            ],
            "operator": "AND",
        },
    )
    print(
        f"DEBUG: Got {len(recently_ended_jobs)} jobs with job[endedAt] >= {starting_timestamp.astimezone().isoformat()}"
    )
    # Filter to jobs where the count of SUCCEEDED tasks is positive.
    recently_ended_jobs = [
        job for job in recently_ended_jobs if job["taskRunStatusCounts"]["SUCCEEDED"] > 0
    ]
    print(f"DEBUG: Filtered down to {len(recently_ended_jobs)} jobs based on SUCCEEDED task filter")
    download_candidate_jobs.update(
        {job["jobId"]: _datetimes_to_str(job) for job in recently_ended_jobs}
    )

    duration = datetime.now(tz=timezone.utc) - start_time
    print_function_callback(f"...retrieval completed in {duration}")

    return download_candidate_jobs


class CategorizedJobIds:
    """
    Takes jobs loaded from a loaded checkpoint and a query to get download candidate jobs,
    analyzes all the jobs by looking at fields like task run status counds to categorize them.

    Job categories:
        added: The job was created or requeued so it now can produce new downloads.
        updated: The job changed since the previous incremental download operation.
        unchanged: The job did not change since the previous incremental download operation.
        completed: The job finished running so all output is available for download.
        inactive: The job can no longer have any new downloads unless it is requeued. Minimal
            metadata is tracked to detect if it is requeued.
        missing_storage_profile: The job has no storage profile, but the operation requires one.
            If incremental download was called with local_storage_profile_id=None, this set
            will always be empty.
        attachments_free: The job has no job attachments associated that can produce
            outputs for download.
    """

    added: set[str] = set()
    updated: set[str] = set()
    unchanged: set[str] = set()
    completed: set[str] = set()
    inactive: set[str] = set()
    missing_storage_profile: set[str] = set()
    attachments_free: set[str] = set()


def _categorize_jobs_in_checkpoint(
    boto3_session: boto3.Session,
    farm_id: str,
    queue_id: str,
    checkpoint: IncrementalDownloadState,
    download_candidate_jobs: dict[str, dict[str, Any]],
    new_completed_timestamp: datetime,
    print_function_callback: Callable[[str], None] = lambda msg: None,
) -> CategorizedJobIds:
    """
    Categorizes the provided download candidate jobs by id into a CategorizedJobIds object,
    updating the jobs within download_candidate_jobs where necessary.

    * Calls boto3 deadline.get_job() to get job attachments manifest information and storage profile id if it is not stored yet.

    Args:
        boto3_session: The boto3.Session for accessing AWS.
        farm_id: The farm id for the operation.
        queue_id: The queue id for the operation.
        checkpoint: The checkpoint for the incremental download.
        download_candidate_jobs: The result of a _get_download_candidate_jobs call, {job_id: job} where
            job is a result from a deadline.search_jobs() or deadline.get_job() call.
        new_completed_timestamp: This is the timestamp value that will be placed in
            checkpoint.downloads_completed_timestamp when saving the checkpoint.
        print_function_callback: Callback for printing output to the terminal or log.
    """
    deadline = get_session_client(boto3_session, "deadline")
    checkpoint_jobs = {job.job_id: job.job for job in checkpoint.jobs}
    checkpoint_job_ids = set(checkpoint_jobs.keys())

    download_candidate_job_ids = set(download_candidate_jobs.keys())

    print_function_callback(
        f"Categorizing {len(checkpoint_jobs)} checkpoint jobs against {len(download_candidate_jobs)} download candidate jobs..."
    )
    start_time = datetime.now(tz=timezone.utc)

    finished_tracking_job_ids = checkpoint_job_ids.difference(download_candidate_job_ids)
    updated_job_ids = checkpoint_job_ids.intersection(download_candidate_job_ids)
    new_job_ids = download_candidate_job_ids.difference(checkpoint_job_ids)
    # The following sets get populated while analyzing the jobs
    unchanged_job_ids = set()
    attachments_free_job_ids = set()
    missing_storage_profile = set()
    completed_job_ids = set()

    # Copy the job attachments manifest data and storage profile id from the checkpoint to the new job objects. This data
    # is not returned by deadline:SearchJobs, so we need to call deadline:GetJob on every job to retrieve it. This data
    # on a job don't change, so after the call to deadline:GetJob we can cache it indefinitely.
    for job_id in updated_job_ids:
        ip_job = checkpoint_jobs[job_id]
        dc_job = download_candidate_jobs[job_id]

        if set(ip_job.keys()) == {"jobId"}:
            # If the job has a minimal placeholder, move the job id to the new job ids
            new_job_ids.add(job_id)
        elif ip_job["attachments"] is None:
            # Carry over the minimal placeholder identifying the job as not using job attachments
            download_candidate_jobs[job_id] = ip_job
            attachments_free_job_ids.add(job_id)
        elif ip_job["storageProfileId"] is None and checkpoint.local_storage_profile_id is not None:
            # Carry over the minimal placeholder identifying the job as missing a storage profile
            download_candidate_jobs[job_id] = ip_job
            missing_storage_profile.add(job_id)
        else:
            # Copy the attachments manifest metadata as it is not returned by deadline:SearchJobs
            dc_job["attachments"] = ip_job["attachments"]
            dc_job["storageProfileId"] = ip_job["storageProfileId"]

    updated_job_ids.difference_update(attachments_free_job_ids)
    updated_job_ids.difference_update(new_job_ids)
    updated_job_ids.difference_update(missing_storage_profile)

    # Prune jobs that we are (almost) certain have no changes by looking at its task status counts. We treat a job as unchanged if its
    # value job["taskRunStatusCounts"]["SUCCEEDED"] stayed the same and its timestamp job["endedAt"] stayed the same.
    #
    # The case this misses (and causes a delay in task output download) is the following sequence: 1/ User requeues one or more steps/tasks.
    # 2/ Tasks succeed in the correct number to equal the previous value 3/ The incremental output download command sees an equal count
    # and miscategorizes it as unchanged. If that count is all the tasks, the job["endedAt"] timestamp will catch it, and if the count
    # is less, the next time a task completes the succeeded count will be different.
    #
    # Because of this potential delay, the checkpoint needs to keep tracking all of the sessions it has seen, and cannot assume
    # that a session ending before the downloads completed timestamp was already processed.
    for job_id in updated_job_ids:
        ip_job = checkpoint_jobs[job_id]
        dc_job = download_candidate_jobs[job_id]

        if ip_job["taskRunStatusCounts"]["SUCCEEDED"] == dc_job["taskRunStatusCounts"][
            "SUCCEEDED"
        ] and ip_job.get("endedAt") == dc_job.get("endedAt"):
            print_function_callback(f"UNCHANGED Job: {dc_job['name']} ({job_id})")
            unchanged_job_ids.add(job_id)
    updated_job_ids.difference_update(unchanged_job_ids)

    # First make note of any jobs that were dropped from tracking, for example if they were canceled or they failed
    for job_id in finished_tracking_job_ids:
        ip_job = checkpoint_jobs[job_id]
        if "taskRunStatusCounts" in ip_job:
            ip_succeeded_task_count = ip_job["taskRunStatusCounts"]["SUCCEEDED"]
            ip_total_task_count = sum(value for _, value in ip_job["taskRunStatusCounts"].items())
        else:
            ip_succeeded_task_count = 0
            ip_total_task_count = -1

        # Print something only if the job is more than a minimal "jobId" tracker
        if set(ip_job.keys()) != {"jobId"}:
            print_function_callback(f"FINISHED TRACKING Job: {ip_job['name']} ({job_id})")
            if ip_job["attachments"] is None:
                print_function_callback("  Job without job attachments is no longer active")
            elif ip_succeeded_task_count == ip_total_task_count:
                print_function_callback("   Job succeeded")
            else:
                print_function_callback(
                    "   Job is not a download candidate anymore (likely suspended, canceled or failed)"
                )

    # Process all the jobs that have updates
    for job_id in updated_job_ids:
        ip_job = checkpoint_jobs[job_id]
        dc_job = download_candidate_jobs[job_id]
        ip_succeeded_task_count = ip_job["taskRunStatusCounts"]["SUCCEEDED"]
        ip_total_task_count = sum(value for _, value in ip_job["taskRunStatusCounts"].items())
        dc_succeeded_task_count = dc_job["taskRunStatusCounts"]["SUCCEEDED"]
        dc_total_task_count = sum(value for _, value in dc_job["taskRunStatusCounts"].items())

        print_function_callback(f"EXISTING Job: {ip_job['name']} ({job_id})")
        print_function_callback(
            f"  Succeeded tasks (before): {ip_succeeded_task_count} / {ip_total_task_count}"
        )
        print_function_callback(
            f"  Succeeded tasks (now)   : {dc_succeeded_task_count} / {dc_total_task_count}"
        )

        # Use the CLI output format to produce a diff of the changes
        ip_job_repr: list[str] = _cli_object_repr(ip_job).splitlines()
        dc_job_repr: list[str] = _cli_object_repr(dc_job).splitlines()

        for line in difflib.unified_diff(
            ip_job_repr,
            dc_job_repr,
            fromfile="Previous update",
            tofile="Current update",
            lineterm="",
        ):
            print_function_callback(f"  {line}")

        if (
            dc_succeeded_task_count == dc_total_task_count
            and "endedAt" in dc_job
            and datetime.fromisoformat(dc_job["endedAt"]) < new_completed_timestamp
        ):
            completed_job_ids.add(job_id)
    updated_job_ids.difference_update(completed_job_ids)

    # Process all the jobs that are new
    for job_id in new_job_ids:
        dc_job = download_candidate_jobs[job_id]

        # Call deadline:GetJob to retrieve attachments manifest information
        job = deadline.get_job(jobId=job_id, queueId=queue_id, farmId=farm_id)
        dc_job["attachments"] = job.get("attachments")
        dc_job["storageProfileId"] = job.get("storageProfileId")
        dc_succeeded_task_count = dc_job["taskRunStatusCounts"]["SUCCEEDED"]
        dc_total_task_count = sum(value for _, value in dc_job["taskRunStatusCounts"].items())

        print_function_callback(f"NEW Job: {dc_job['name']} ({job_id})")

        if (
            dc_job["attachments"] is not None
            and dc_job["storageProfileId"] is None
            and checkpoint.local_storage_profile_id is not None
        ):
            print_function_callback(
                "  WARNING: THE JOB OUTPUT WILL NOT BE DOWNLOADED, IT HAS NO STORAGE PROFILE."
            )
            missing_storage_profile.add(job_id)
            continue

        print_function_callback(
            f"  Succeeded tasks: {dc_succeeded_task_count} / {dc_total_task_count}"
        )
        if dc_job["attachments"] is None:
            # If the job does not use job attachments, save a minimal placeholder to avoid
            # repeatedly calling deadline:GetJob.
            download_candidate_jobs[job_id] = dc_job = {
                "jobId": job_id,
                "name": dc_job["name"],
                "attachments": None,
            }
            attachments_free_job_ids.add(job_id)
            print_function_callback("  Job does not use job attachments.")
        else:
            print_function_callback("  Manifest file system paths:")
            for manifest in dc_job["attachments"]["manifests"]:
                print_function_callback(
                    f"    - {manifest['rootPath']} ({manifest['rootPathFormat']})"
                )

        if (
            dc_succeeded_task_count == dc_total_task_count
            and "endedAt" in dc_job
            and datetime.fromisoformat(dc_job["endedAt"]) < new_completed_timestamp
        ):
            completed_job_ids.add(job_id)
    new_job_ids.difference_update(attachments_free_job_ids)
    new_job_ids.difference_update(completed_job_ids)
    new_job_ids.difference_update(missing_storage_profile)

    result = CategorizedJobIds()
    result.attachments_free = attachments_free_job_ids
    result.missing_storage_profile = missing_storage_profile
    result.completed = completed_job_ids
    result.inactive = finished_tracking_job_ids
    result.added = new_job_ids
    result.unchanged = unchanged_job_ids
    result.updated = updated_job_ids

    duration = datetime.now(tz=timezone.utc) - start_time
    print_function_callback(f"...categorization completed in {duration}")

    return result


def _retrieve_sessions_for_job(
    deadline_client: BaseClient,
    checkpoint: IncrementalDownloadState,
    farm_id: str,
    queue_id: str,
    job_id: str,
    session_ended_threshold: datetime,
    output_job_sessions: dict[str, list],
):
    """
    Uses deadline.list_sessions to get all sessions of the specified job that are still running or
    that ended after session_ended_threshold.

    Places the output into output_job_sessions[job_id]

    Args:
        deadline_client: A boto3 client for accessing Deadline.
        checkpoint: The checkpoint for the incremental download.
        farm_id: The farm id for the operation.
        queue_id: The queue id for the operation.
        job_id: The job id to process.
        session_ended_threshold: The timestamp threshold to filter out older sessions based on the endedAt field.
        output_job_sessions: A dictionary {job_id: session_list} to populate for the provided job id.
    """
    sessions_paginator = deadline_client.get_paginator("list_sessions")
    # Filter out older sessions by endedAt timestamp, using an eventual consistency window to accept a little extra
    session_ended_threshold = session_ended_threshold - timedelta(
        seconds=checkpoint.eventual_consistency_max_seconds
    )

    session_list: list[dict[str, Any]] = []
    for sessions_page in sessions_paginator.paginate(
        farmId=farm_id, queueId=queue_id, jobId=job_id
    ):
        for session in sessions_page.get("sessions", []):
            if "endedAt" not in session or session["endedAt"] >= session_ended_threshold:
                session_list.append(session)
    if session_list:
        output_job_sessions[job_id] = session_list


def _retrieve_session_actions_for_session(
    deadline_client: BaseClient,
    checkpoint_job_session_completed_indexes: dict[str, dict[str, int]],
    farm_id: str,
    queue_id: str,
    job_id: str,
    output_session: dict[str, Any],
):
    """
    Args:
        deadline_client: A boto3 client for accessing Deadline.
        checkpoint_job_session_completed_indexes: All the jobs' session action indexes loaded from the checkpoint.
            The value checkpoint_job_session_completed_indexes[job_id][session_id] is the session action index of
            the latest session action that is completed download.
        farm_id: The farm id for the operation.
        queue_id: The queue id for the operation.
        job_id: The job id to process.
        output_session: The session to populate with a sessionActions field.
    """
    session_actions_paginator = deadline_client.get_paginator("list_session_actions")

    session_action_list: list[dict[str, Any]] = []
    for session_actions_page in session_actions_paginator.paginate(
        farmId=farm_id,
        queueId=queue_id,
        jobId=job_id,
        sessionId=output_session["sessionId"],
    ):
        # Include only succeeded taskRun actions.
        for session_action in session_actions_page.get("sessionActions", []):
            succeeded = session_action.get("status") == "SUCCEEDED"
            is_task_run = "taskRun" in session_action.get("definition", {})
            if succeeded and is_task_run:
                session_action_list.append(session_action)

    if session_action_list:
        # Extract the session action indexes from the ids
        for session_action in session_action_list:
            # Session action IDs look like "sessionaction-abc123-12" for index 12
            session_action_index = int(session_action["sessionActionId"].rsplit("-", 1)[-1])
            session_action["sessionActionIndex"] = session_action_index
        # Include only session action indexes newer than latest downloaded ones from the checkpoint
        session_completed_index: Optional[int] = checkpoint_job_session_completed_indexes.get(
            job_id, {}
        ).get(output_session["sessionId"])
        if session_completed_index is not None:
            # Filter out older session actions that were already downloaded
            session_action_list = [
                session_action
                for session_action in session_action_list
                if session_action["sessionActionIndex"] > session_completed_index
            ]
        if session_action_list:
            output_session["sessionActions"] = session_action_list


def _get_job_sessions(
    boto3_session: boto3.Session,
    boto3_session_for_s3: boto3.Session,
    farm_id: str,
    queue: dict[str, Any],
    checkpoint_job_session_completed_indexes: dict[str, dict[str, int]],
    categorized_job_ids: CategorizedJobIds,
    checkpoint: IncrementalDownloadState,
    download_candidate_jobs: dict[str, dict[str, Any]],
    print_function_callback: Callable[[str], None] = lambda msg: None,
) -> dict[str, list]:
    """
    This function gets all the job sessions and session actions from the completed, added, and updated jobs.
    It uses the checkpoint's session_completed_indexes to filter out older session actions that are already downloaded.

    Args:
        boto3_session: The boto3.Session for accessing AWS.
        boto3_session_for_s3: The boto3.Session to use for accessing S3.
        farm_id: The farm id for the operation.
        queue: The queue as returned by boto3 deadline.get_queue().
        checkpoint_job_session_completed_indexes: All the jobs' session action indexes loaded from the checkpoint.
            The value checkpoint_job_session_completed_indexes[job_id][session_id] is the session action index of
            the latest session action that is completed download.
        categorized_job_ids: The categorized job ids as returned by _categorize_jobs_in_checkpoint().
        checkpoint: The checkpoint for the incremental download.
        download_candidate_jobs: The result of a _get_download_candidate_jobs call, {job_id: job} where
            job is a result from a deadline.search_jobs() or deadline.get_job() call.
        print_function_callback: Callback for printing output to the terminal or log.

    Returns:
        Access a session action in the returned job_sessions with
            job_sessions[job_id][session_index]["sessionActions"][session_action_index]
        The returned structure looks like this:
        {
            "<job_id>": [
                {
                    "sessionId": "<session_id>",
                    ...,
                    "sessionActions": [
                        {
                            "sessionActionId": "<session_action_id>",
                            ...
                        },
                        ...
                    ]
                },
                ...
            ],
            ...
        }
    """
    job_ids = categorized_job_ids.completed.union(categorized_job_ids.added).union(
        categorized_job_ids.updated
    )
    print_function_callback(f"Retrieving sessions for {len(job_ids)} jobs...")
    start_time = datetime.now(tz=timezone.utc)

    # The max timestamp of a downloaded session's endedAt provides a lower bound to filter sessions by.
    # This is tracked in the checkpoint.
    job_session_ended_timestamp: dict[str, datetime] = {
        job.job_id: job.session_ended_timestamp
        for job in checkpoint.jobs
        if job.session_ended_timestamp is not None
    }

    deadline = get_session_client(boto3_session, "deadline")
    job_sessions: dict[str, list] = {}

    # Retrieve all the sessions with some parallelism
    max_workers = SESSIONS_API_MAX_CONCURRENCY
    print_function_callback(f"Using {max_workers} threads")
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for job_id in job_ids:
            # Use the greater of the bootstrap command timestamp and the session ended timestamps
            # recorded in the checkpoint.
            session_ended_threshold = job_session_ended_timestamp.get(job_id)
            if session_ended_threshold is None:
                session_ended_threshold = checkpoint.downloads_started_timestamp

            futures.append(
                executor.submit(
                    _retrieve_sessions_for_job,
                    deadline,
                    checkpoint,
                    farm_id,
                    queue["queueId"],
                    job_id,
                    session_ended_threshold,
                    job_sessions,
                )
            )

        # surfaces any exceptions in the thread
        for future in concurrent.futures.as_completed(futures):
            future.result()

    duration = datetime.now(tz=timezone.utc) - start_time
    print_function_callback(f"...retrieval completed in {duration}")

    print_function_callback("")
    print_function_callback(
        f"Retrieving session actions for {sum(len(session_list) for session_list in job_sessions.values())} sessions..."
    )
    start_time = datetime.now(tz=timezone.utc)

    # Retrieve all the session actions with some parallelism
    max_workers = SESSIONS_API_MAX_CONCURRENCY
    print_function_callback(f"Using {max_workers} threads")
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for job_id, session_list in job_sessions.items():
            for session in session_list:
                futures.append(
                    executor.submit(
                        _retrieve_session_actions_for_session,
                        deadline,
                        checkpoint_job_session_completed_indexes,
                        farm_id,
                        queue["queueId"],
                        job_id,
                        session,
                    )
                )
        # surfaces any exceptions in the thread
        for future in concurrent.futures.as_completed(futures):
            future.result()

    duration = datetime.now(tz=timezone.utc) - start_time
    print_function_callback(f"...retrieval completed in {duration}")

    print_function_callback("")
    print_function_callback("Populating missing manifest S3 keys...")
    start_time = datetime.now(tz=timezone.utc)

    _add_missing_output_manifests_to_job_sessions(
        boto3_session_for_s3, farm_id, queue, job_sessions, download_candidate_jobs
    )

    _filter_session_actions_without_manifests_from_job_sessions(
        job_sessions,
        download_candidate_jobs,
        print_function_callback,
    )

    duration = datetime.now(tz=timezone.utc) - start_time
    print_function_callback(f"...populated in {duration}")

    return job_sessions


def _get_storage_profiles(
    deadline: BaseClient,
    farm_id: str,
    queue: dict[str, Any],
    job_sessions: dict[str, list],
    checkpoint: IncrementalDownloadState,
    download_candidate_jobs: dict[str, dict[str, Any]],
) -> dict[str, dict[str, Any]]:
    """
    Retrieves all the needed storage profiles. To call this function,
    checkpoint.local_storage_profile_id must be set.

    Args:
        deadline: An AWS client for calling deadline APIs.
        farm_id: The farm id for the operation.
        queue: The queue as returned by boto3 deadline.get_queue().
        checkpoint: The checkpoint for the incremental download.
        job_sessions: Contains each job's sessions and session actions, structured as job_sessions[job_id][session_index]["sessionActions"][session_action_index].
                      See the function _get_job_sessions for more details.
        download_candidate_jobs: The result of a _get_download_candidate_jobs call, {job_id: job} where
            job is a result from a deadline.search_jobs() or deadline.get_job() call.
    """
    if checkpoint.local_storage_profile_id is None:
        raise ValueError("The checkpoint local storage profile id must be set.")

    # Collect all the storage profile ids
    storage_profile_ids: set[str] = {checkpoint.local_storage_profile_id}
    for job_id in job_sessions.keys():
        storage_profile_ids.add(download_candidate_jobs[job_id]["storageProfileId"])

    # Load all the storage profiles from Deadline Cloud
    storage_profiles = {
        storage_profile_id: deadline.get_storage_profile_for_queue(
            farmId=farm_id,
            queueId=queue["queueId"],
            storageProfileId=storage_profile_id,
        )
        for storage_profile_id in storage_profile_ids
    }

    return storage_profiles


def _create_path_mapping_rule_appliers(
    storage_profiles: dict[str, dict[str, Any]],
    checkpoint: IncrementalDownloadState,
    download_candidate_jobs: dict[str, dict[str, Any]],
    print_function_callback: Callable[[str], None] = lambda msg: None,
) -> dict[str, Optional[_PathMappingRuleApplier]]:
    """
    Retrieves all the needed storage profiles and constructs path mapping rule applies for them.
    To call this function, checkpoint.local_storage_profile_id must be set.

    Args:
        storage_profiles: A mapping from storage profile id to the storage profile as returned by boto3 deadline.get_storage_profile_for_queue.
        checkpoint: The checkpoint for the incremental download.
        job_sessions: Contains each job's sessions and session actions, structured as job_sessions[job_id][session_index]["sessionActions"][session_action_index].
                      See the function _get_job_sessions for more details.
        download_candidate_jobs: The result of a _get_download_candidate_jobs call, {job_id: job} where
            job is a result from a deadline.search_jobs() or deadline.get_job() call.
        print_function_callback: Callback for printing output to the terminal or log.
    """
    if checkpoint.local_storage_profile_id is None:
        raise ValueError("The checkpoint local storage profile id must be set.")

    path_mapping_rule_appliers: dict[str, Optional[_PathMappingRuleApplier]] = {}

    # Create a path mapping rule applier for each storage profile
    local_storage_profile = storage_profiles[checkpoint.local_storage_profile_id]
    local_storage_profile_name = local_storage_profile["displayName"]
    print_function_callback("")
    print_function_callback(
        f"Local storage profile is {local_storage_profile_name} ({checkpoint.local_storage_profile_id})"
    )
    print_function_callback(
        f"  {len([job for job in download_candidate_jobs.values() if job.get('storageProfileId') == checkpoint.local_storage_profile_id])} download candidate jobs have the same storage profile and will be downloaded to their original specified paths"
    )
    for storage_profile_id, storage_profile in storage_profiles.items():
        storage_profile_name = storage_profile["displayName"]
        if storage_profile_id == checkpoint.local_storage_profile_id:
            path_mapping_rule_appliers[storage_profile_id] = None
        else:
            rules = _generate_path_mapping_rules(storage_profile, local_storage_profile)
            path_mapping_rule_appliers[storage_profile_id] = _PathMappingRuleApplier(rules)

            # Print the path mapping rules for each source storage profile
            print_function_callback("")
            job_count = len(
                [
                    job
                    for job in download_candidate_jobs.values()
                    if job.get("storageProfileId") == storage_profile_id
                ]
            )
            print_function_callback(
                f"Path mapping rules for {job_count} download candidate jobs with storage profile {storage_profile_name} ({storage_profile_id})"
            )
            print_function_callback(
                f"  job storage profile: {storage_profile_name} ({storage_profile['osFamily']})"
            )
            print_function_callback(
                f"  local storage profile: {local_storage_profile_name} ({local_storage_profile['osFamily']})"
            )
            if rules:
                for rule in rules:
                    print_function_callback(f"  - from: {rule.source_path}")
                    print_function_callback(f"    to:   {rule.destination_path}")
            else:
                print_function_callback(
                    f"   No rules generated. Storage profiles {local_storage_profile_name} and {storage_profile_name} share no file system location names."
                )
    return path_mapping_rule_appliers


def _add_missing_output_manifests_to_job_sessions(
    boto3_session_for_s3: boto3.Session,
    farm_id: str,
    queue: dict[str, Any],
    job_sessions: dict[str, list],
    download_candidate_jobs: dict[str, dict[str, Any]],
):
    """
    Args:
        boto3_session_for_s3: The boto3.Session to use for accessing S3.
        farm_id: The farm id for the operation.
        queue: The queue as returned by boto3 deadline.get_queue().
        job_sessions: Contains each job's sessions and session actions, structured as job_sessions[job_id][session_index]["sessionActions"][session_action_index].
                      See the function _get_job_sessions for more details.
        download_candidate_jobs: The result of a _get_download_candidate_jobs call, {job_id: job} where
            job is a result from a deadline.search_jobs() or deadline.get_job() call.
    """
    for job_id, session_list in job_sessions.items():
        job = download_candidate_jobs[job_id]
        session_action_list = [
            session_action
            for session in session_list
            for session_action in session.get("sessionActions", [])
        ]
        _add_output_manifests_from_s3(
            farm_id, queue, job, boto3_session_for_s3, session_action_list
        )


def _filter_session_actions_without_manifests_from_job_sessions(
    job_sessions: dict[str, list],
    download_candidate_jobs: dict[str, dict[str, Any]],
    print_function_callback: Callable[[str], None] = lambda msg: None,
):
    """
    Modify job_sessions in place to filter out any session actions that lack any output manifests.
    Print a warning message for any job that had a session action like this.

    Args:
        job_sessions: Contains each job's sessions and session actions, structured as job_sessions[job_id][session_index]["sessionActions"][session_action_index].
                      See the function _get_job_sessions for more details.
        download_candidate_jobs: The result of a _get_download_candidate_jobs call, {job_id: job} where
            job is a result from a deadline.search_jobs() or deadline.get_job() call.
        print_function_callback: Callback for printing output to the terminal or log.
    """
    for job_id, session_list in job_sessions.items():
        job = download_candidate_jobs[job_id]
        total_count = 0
        filtered_count = 0
        for session in session_list:
            total_count += len(session.get("sessionActions", []))
            # Filter out session actions with no manifest files
            filtered_session_action_list = [
                session_action
                for session_action in session.get("sessionActions", [])
                if any(item != {} for item in session_action["manifests"])
            ]
            filtered_count += len(filtered_session_action_list)
            if total_count != filtered_count:
                session["sessionActions"] = filtered_session_action_list
        if total_count != filtered_count:
            print_function_callback(
                f"WARNING: Job {job['name']} ({job_id}) ran {total_count - filtered_count} / {total_count} session actions with no output."
            )
            print_function_callback(
                "         This may indicate steps in the job that strictly perform validation or save results elsewhere like a shared file system or S3."
            )


def _update_checkpoint_jobs_list(
    checkpoint: IncrementalDownloadState,
    download_candidate_jobs: dict[str, dict[str, Any]],
    categorized_job_ids: CategorizedJobIds,
    job_sessions: dict[str, list],
):
    """
    Update the jobs list in the checkpoint object.

    Args:
        checkpoint: The checkpoint for the incremental download.
        download_candidate_jobs: The result of a _get_download_candidate_jobs call, {job_id: job} where
            job is a result from a deadline.search_jobs() or deadline.get_job() call.
        categorized_job_ids: The categorized job ids as returned by _categorize_jobs_in_checkpoint().
        job_sessions: Contains each job's sessions and session actions, structured as job_sessions[job_id][session_index]["sessionActions"][session_action_index].
                      See the function _get_job_sessions for more details.
    """
    updated_jobs: list[IncrementalDownloadJob] = []

    # Produce the session_ended_timestamp for all the job ids. Start
    # with the values from the previous checkpoint, and then overwrite
    # them from job_sessions
    job_session_ended_timestamps: dict[str, Optional[datetime]] = {
        job.job_id: job.session_ended_timestamp
        for job in checkpoint.jobs
        if job.session_ended_timestamp is not None
    }
    for job_id, session_list in job_sessions.items():
        max_session_ended_timestamp = None
        for session in session_list:
            if "endedAt" in session:
                if max_session_ended_timestamp is None:
                    max_session_ended_timestamp = session["endedAt"]
                else:
                    max_session_ended_timestamp = max(
                        max_session_ended_timestamp, session["endedAt"]
                    )
        job_session_ended_timestamps[job_id] = max_session_ended_timestamp

    # Produce the session_completed_indexes for all the job ids. Start
    # with the values from the previous checkpoint, then overwrite
    # them from job_sessions.
    job_session_completed_indexes: dict[str, dict[str, int]] = {
        job.job_id: job.session_completed_indexes for job in checkpoint.jobs
    }
    for job_id, session_list in job_sessions.items():
        for session in session_list:
            session_actions = session.get("sessionActions", [])
            if session_actions:
                job_session_completed_indexes.setdefault(job_id, {})[session["sessionId"]] = max(
                    session_action["sessionActionIndex"] for session_action in session_actions
                )
        job_session_ended_timestamps[job_id] = max_session_ended_timestamp

    # These categories keep the download_candidate_jobs job as is.
    for job_id in (
        categorized_job_ids.added | categorized_job_ids.updated | categorized_job_ids.unchanged
    ):
        updated_jobs.append(
            IncrementalDownloadJob(
                download_candidate_jobs[job_id],
                job_session_ended_timestamps.get(job_id),
                job_session_completed_indexes.get(job_id, {}),
            )
        )
    # This category keeps a signal that it has no job attachments to process by having an attachments field with None in it
    for job_id in categorized_job_ids.attachments_free:
        updated_jobs.append(
            IncrementalDownloadJob(
                {
                    "jobId": job_id,
                    "name": download_candidate_jobs[job_id]["name"],
                    "attachments": None,
                },
                None,
                {},
            )
        )
    # This category keeps a signal that it is missing a storage profile by populating the attachments but
    # having a storageProfileId field with None in it. By keeping the attachments in the checkpoint, someone
    # inspecting the checkpoint to understand what's happening can get an idea about the paths for the job
    # and diagnose problems quicker.
    for job_id in categorized_job_ids.missing_storage_profile:
        updated_jobs.append(
            IncrementalDownloadJob(
                {
                    "jobId": job_id,
                    "name": download_candidate_jobs[job_id]["name"],
                    "attachments": download_candidate_jobs[job_id]["attachments"],
                    "storageProfileId": None,
                },
                None,
                {},
            )
        )
    # Keep completed jobs around until they become inactive
    for job_id in categorized_job_ids.completed:
        updated_jobs.append(
            IncrementalDownloadJob(
                download_candidate_jobs[job_id],
                job_session_ended_timestamps.get(job_id),
                job_session_completed_indexes.get(job_id, {}),
            )
        )
    # When a job becomes inactive, keep it around in minimal form when it has a session_ended_timestamp.
    # This is necessary for the case where a completed job gets requeued later. We can't tell
    # that it was requeued from the deadline.search_jobs query, so we hold this metadata in the checkpoint.
    for job_id in categorized_job_ids.inactive:
        session_ended_timestamp = job_session_ended_timestamps.get(job_id)
        if session_ended_timestamp is not None:
            updated_jobs.append(
                IncrementalDownloadJob({"jobId": job_id}, session_ended_timestamp, {})
            )

    checkpoint.jobs = updated_jobs


@api.record_function_latency_telemetry_event()
def _incremental_output_download(
    farm_id: str,
    queue: dict[str, Any],
    boto3_session: boto3.Session,
    checkpoint: IncrementalDownloadState,
    file_conflict_resolution: FileConflictResolution,
    config: Optional[ConfigParser] = None,
    print_function_callback: Callable[[str], None] = lambda msg: None,
    *,
    dry_run: bool = False,
) -> IncrementalDownloadState:
    """
    This function downloads all the task run outputs from the specified queue, that have become
    available since the last time the function was called. The checkpoint object
    keeps track of all state needed to keep track of what needs to be downloaded.

    Pre-condition: The input checkpoint holds all information needed to understand the state of downloads
        completed up to the timestamp checkpoint.downloads_completed_timestamp. See the documentation
        in the IncrementalDownloadState to understand the invariants of the checkpoint.

    Post-condition: The output checkpoint has an updated checkpoint.downloads_completed_timestamp,
        all downloads were performed up to at least this timestamp, and the checkpoint data
        is updated to satisfy the next call's pre-condition.

    Args:
        farm_id: The farm id for the operation.
        queue: The queue as returned by boto3 deadline.get_queue().
        boto3_session: The boto3.Session for accessing AWS.
        checkpoint: The checkpoint for the incremental download.
        config: Optional, a Deadline Cloud configuration as loaded from config_file.read_config().
        print_function_callback: Callback for printing output to the terminal or log.
        dry_run: If True, the operation will print out information but not perform any data downloads.

    Returns:
        An updated checkpoint object.
    """
    if sys.version_info < (3, 9):
        raise DeadlineOperationError("The sync-output command requires Python version 3.9 or later")

    durations = IncrementalOutputDownloadLatencies()
    deadline = get_session_client(boto3_session, "deadline")

    # When this function is done, we will be confident that downloads are complete up to
    # new_completed_timestamp. We subtract a duration from now() that gives a generous amount of
    # time for the deadline:SearchJobs API's eventual consistency to converge.
    current_timestamp = datetime.now(timezone.utc)
    new_completed_timestamp = max(
        checkpoint.downloads_started_timestamp,
        current_timestamp - timedelta(seconds=checkpoint.eventual_consistency_max_seconds),
    )

    # The queue role is used for accessing S3
    boto3_session_for_s3 = api.get_queue_user_boto3_session(
        deadline=deadline,
        config=config,
        farm_id=farm_id,
        queue_id=queue["queueId"],
        queue_display_name=queue["displayName"],
    )

    print_function_callback("Updating download state across time interval:")
    print_function_callback(
        f"    From: {checkpoint.downloads_completed_timestamp.astimezone().isoformat()}"
    )
    print_function_callback(f"      To: {current_timestamp.astimezone().isoformat()}")
    update_length = current_timestamp - checkpoint.downloads_completed_timestamp
    eventual_consistency_delta = timedelta(seconds=checkpoint.eventual_consistency_max_seconds)
    if update_length > eventual_consistency_delta:
        print_function_callback(
            f"  Length: {update_length - eventual_consistency_delta} + {eventual_consistency_delta} (eventual consistency allowance)"
        )
    else:
        # Immediately after bootstrapping, this length will be shorter than the eventual consistency window
        print_function_callback(f"  Length: {update_length}")
    print_function_callback("")

    # Save all the jobs' session action indexes from the checkpoint, before we update the checkpoint's jobs list
    checkpoint_job_session_completed_indexes: dict[str, dict[str, int]] = {
        job.job_id: job.session_completed_indexes for job in checkpoint.jobs
    }

    # Call deadline:SearchJobs to get a set of jobs that includes every job with downloads available.
    start_t = time.perf_counter_ns()
    download_candidate_jobs: dict[str, dict[str, Any]] = _get_download_candidate_jobs(
        boto3_session,
        farm_id,
        queue["queueId"],
        checkpoint.downloads_completed_timestamp,
        print_function_callback,
    )
    durations._get_download_candidate_jobs = time.perf_counter_ns() - start_t

    print_function_callback("")

    # Compare the download candidates with the previously saved checkpoint state to categorize the jobs
    start_t = time.perf_counter_ns()
    categorized_job_ids: CategorizedJobIds = _categorize_jobs_in_checkpoint(
        boto3_session,
        farm_id,
        queue["queueId"],
        checkpoint,
        download_candidate_jobs,
        new_completed_timestamp,
        print_function_callback,
    )
    durations._categorize_jobs_in_checkpoint = time.perf_counter_ns() - start_t

    print_function_callback("")

    # All the completed, added, and updated jobs might have downloads available. Retrieve the sessions for these jobs.
    start_t = time.perf_counter_ns()
    job_sessions: dict[str, list] = _get_job_sessions(
        boto3_session,
        boto3_session_for_s3,
        farm_id,
        queue,
        checkpoint_job_session_completed_indexes,
        categorized_job_ids,
        checkpoint,
        download_candidate_jobs,
        print_function_callback,
    )
    durations._get_job_sessions = time.perf_counter_ns() - start_t

    # If storage profiles are being used, get them and construct all the path mapping rules
    path_mapping_rule_appliers: dict[str, Optional[_PathMappingRuleApplier]] = {}
    if checkpoint.local_storage_profile_id:
        start_t = time.perf_counter_ns()
        storage_profiles = _get_storage_profiles(
            deadline, farm_id, queue, job_sessions, checkpoint, download_candidate_jobs
        )
        path_mapping_rule_appliers = _create_path_mapping_rule_appliers(
            storage_profiles,
            checkpoint,
            download_candidate_jobs,
            print_function_callback,
        )
        durations.path_mapping = time.perf_counter_ns() - start_t

    # Use the information collected so far to update the jobs list in checkpoint
    start_t = time.perf_counter_ns()
    _update_checkpoint_jobs_list(
        checkpoint, download_candidate_jobs, categorized_job_ids, job_sessions
    )
    durations._update_checkpoint_jobs_list = time.perf_counter_ns() - start_t

    start_t = time.perf_counter_ns()
    unmapped_paths: dict[str, list[str]] = {}
    downloaded_manifests: list[tuple[datetime, BaseAssetManifest]] = (
        _download_all_manifests_with_absolute_paths(
            queue,
            download_candidate_jobs,
            job_sessions,
            path_mapping_rule_appliers,
            unmapped_paths,
            boto3_session_for_s3,
            print_function_callback,
        )
    )
    durations._download_all_manifests_with_absolute_paths = time.perf_counter_ns() - start_t

    # Print warning messages about all the output paths that will not be downloaded due to lack of path mapping.
    if unmapped_paths:
        print_function_callback("")
        print_function_callback("WARNING: THE FOLLOWING FILES WILL NOT BE DOWNLOADED")
        for job_id, unmapped_path_list in unmapped_paths.items():
            print_function_callback(
                f"    Job {download_candidate_jobs[job_id]['name']} ({job_id}) has outputs with unmapped paths that will not be downloaded"
            )
            storage_profile = storage_profiles[download_candidate_jobs[job_id]["storageProfileId"]]
            print_function_callback(
                f"      Job storage profile is {storage_profile['displayName']} ({storage_profile['storageProfileId']})"
            )
            print_function_callback("      Summary of unmapped paths:")
            path_format = (
                PathFormat.WINDOWS
                if storage_profile["osFamily"] == StorageProfileOperatingSystemFamily.WINDOWS.value
                else PathFormat.POSIX
            )
            paths_summary = summarize_path_list(
                unmapped_path_list, max_entries=30, path_format=path_format
            )
            print_function_callback(textwrap.indent(paths_summary, "      "))

    # Merge the manifests ordered by the last modified timestamp
    manifest_paths_to_download: list[BaseManifestPath] = _merge_absolute_path_manifest_list(
        downloaded_manifests
    )

    # Print a summary of all the paths before starting the download
    local_path_list = [manifest_path.path for manifest_path in manifest_paths_to_download]
    file_size_by_path = {
        manifest_path.path: manifest_path.size for manifest_path in manifest_paths_to_download
    }
    print_function_callback("")
    print_function_callback("Summary of paths to download:")
    print_function_callback(
        summarize_path_list(local_path_list, total_size_by_path=file_size_by_path, max_entries=30)
    )
    print_function_callback("")

    if not dry_run:
        print_function_callback(f"Downloading {len(manifest_paths_to_download)} files from S3...")
        start_t = time.perf_counter_ns()
        start_time = datetime.now(tz=timezone.utc)

        # Incremental download is mostly a background thing, so don't print status too often while downloading
        MIN_DELAY_BETWEEN_PRINTOUTS = 20
        last_call_time = time.time() - MIN_DELAY_BETWEEN_PRINTOUTS
        printed_100_percent = False

        def _update_download_progress(
            download_metadata: ProgressReportMetadata,
        ) -> bool:
            nonlocal last_call_time, printed_100_percent

            if not printed_100_percent and download_metadata.progress == 100:
                print_function_callback(f"{download_metadata.progressMessage}")
                last_call_time = time.time()
                printed_100_percent = True
            elif (
                not printed_100_percent
                and time.time() - last_call_time > MIN_DELAY_BETWEEN_PRINTOUTS
            ):
                print_function_callback(f"{download_metadata.progressMessage}")
                last_call_time = time.time()

            return sigint_handler.continue_operation

        _download_manifest_paths(
            manifest_paths_to_download,
            HashAlgorithm.XXH128,
            queue,
            boto3_session_for_s3,
            file_conflict_resolution,
            on_downloading_files=_update_download_progress,
            print_function_callback=print_function_callback,
        )

        durations.download = time.perf_counter_ns() - start_t
        duration = datetime.now(tz=timezone.utc) - start_time
        print_function_callback(f"...downloaded in {duration}")
    else:
        print_function_callback("Skipping downloads due to DRY RUN")

    # Update the timestamp in the state object to reflect the downloads that were completed
    checkpoint.downloads_completed_timestamp = new_completed_timestamp

    stats: dict[str, Any] = {
        "downloaded_session_actions": sum(
            len(session.get("sessionActions", []))
            for session_list in job_sessions.values()
            for session in session_list
        ),
        "downloaded_files": len(manifest_paths_to_download),
        "downloaded_bytes": sum(path.size for path in manifest_paths_to_download),
        "jobs_with_downloads": {
            "completed": len(categorized_job_ids.completed),
            "added": len(categorized_job_ids.added),
            "updated": len(categorized_job_ids.updated),
        },
        "jobs_without_downloads": {
            "not_using_job_attachments": len(categorized_job_ids.attachments_free),
            "missing_storage_profile": len(categorized_job_ids.missing_storage_profile),
            "unchanged": len(categorized_job_ids.unchanged),
            "inactive": len(categorized_job_ids.inactive),
        },
        "unmapped_paths": len(unmapped_paths),
    }
    api.get_deadline_cloud_library_telemetry_client().record_event(
        event_type="com.amazon.rum.deadline.queue_sync_output_stats",
        event_details={
            "latencies": asdict(durations),
            "dry_run": dry_run,
            **stats,
        },
    )

    print_function_callback("")
    if dry_run:
        print_function_callback(
            "Summary of DRY RUN for incremental output download (no files were downloaded to the file system):"
        )
    else:
        print_function_callback("Summary of incremental output download:")
    print_function_callback(f"  Downloaded session actions: {stats['downloaded_session_actions']}")
    print_function_callback(f"  Downloaded files: {stats['downloaded_files']}")
    print_function_callback(
        f"  Downloaded bytes: {human_readable_file_size(stats['downloaded_bytes'])}"
    )
    print_function_callback("  Jobs with downloads:")
    print_function_callback(f"    completed: {stats['jobs_with_downloads']['completed']}")
    print_function_callback(f"    added: {stats['jobs_with_downloads']['added']}")
    print_function_callback(f"    updated: {stats['jobs_with_downloads']['updated']}")
    print_function_callback("  Jobs without downloads:")
    print_function_callback(
        f"    not using job attachments: {stats['jobs_without_downloads']['not_using_job_attachments']}"
    )
    print_function_callback(
        f"    missing storage profile: {stats['jobs_without_downloads']['missing_storage_profile']}"
    )
    print_function_callback(f"    unchanged: {stats['jobs_without_downloads']['unchanged']}")
    print_function_callback(f"    inactive: {stats['jobs_without_downloads']['inactive']}")

    return checkpoint



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from ._session import (
    get_boto3_client,
    get_user_and_identity_store_id,
)
from .. import api


def _call_paginated_deadline_list_api(list_api, list_property_name, **kwargs):
    """
    Calls a deadline:List* API repeatedly to concatenate all pages.

    Example:
        deadline = get_boto3_client("deadline")
        return _call_paginated_deadline_list_api(deadline.list_farms, "farms", **kwargs)

    Args:
      list_api (callable): The List* API function to call, from the boto3 client.
      list_property_name (str): The name of the property in the response that contains
                                the list.
    """
    response = list_api(**kwargs)
    result = {list_property_name: response[list_property_name]}

    while "nextToken" in response:
        response = list_api(nextToken=response["nextToken"], **kwargs)
        result[list_property_name].extend(response[list_property_name])

    return result


@api.record_function_latency_telemetry_event()
def list_farms(config=None, **kwargs):
    """
    Calls the deadline:ListFarms API call, applying the filter for user membership
    depending on the configuration. If the response is paginated, it repeated
    calls the API to get all the farms.
    """
    if "principalId" not in kwargs:
        user_id, _ = get_user_and_identity_store_id(config=config)
        if user_id:
            kwargs["principalId"] = user_id

    deadline = get_boto3_client("deadline", config=config)
    return _call_paginated_deadline_list_api(deadline.list_farms, "farms", **kwargs)


@api.record_function_latency_telemetry_event()
def list_queues(config=None, **kwargs):
    """
    Calls the deadline:ListQueues API call, applying the filter for user membership
    depending on the configuration. If the response is paginated, it repeated
    calls the API to get all the queues.
    """
    if "principalId" not in kwargs:
        user_id, _ = get_user_and_identity_store_id(config=config)
        if user_id:
            kwargs["principalId"] = user_id

    deadline = get_boto3_client("deadline", config=config)
    return _call_paginated_deadline_list_api(deadline.list_queues, "queues", **kwargs)


@api.record_function_latency_telemetry_event()
def list_jobs(config=None, **kwargs):
    """
    Calls the deadline:ListJobs API call, applying the filter for user membership
    depending on the configuration. If the response is paginated, it repeated
    calls the API to get all the jobs.
    """
    if "principalId" not in kwargs:
        user_id, _ = get_user_and_identity_store_id(config=config)
        if user_id:
            kwargs["principalId"] = user_id

    deadline = get_boto3_client("deadline", config=config)
    return _call_paginated_deadline_list_api(deadline.list_jobs, "jobs", **kwargs)


@api.record_function_latency_telemetry_event()
def list_fleets(config=None, **kwargs):
    """
    Calls the deadline:ListFleets API call, applying the filter for user membership
    depending on the configuration. If the response is paginated, it repeated
    calls the API to get all the fleets.
    """
    if "principalId" not in kwargs:
        user_id, _ = get_user_and_identity_store_id(config=config)
        if user_id:
            kwargs["principalId"] = user_id

    deadline = get_boto3_client("deadline", config=config)
    return _call_paginated_deadline_list_api(deadline.list_fleets, "fleets", **kwargs)


@api.record_function_latency_telemetry_event()
def list_storage_profiles_for_queue(config=None, **kwargs):
    """
    Calls the deadline:ListStorageProfilesForQueue API call, applying the filter for user membership
    depending on the configuration. If the response is paginated, it repeated
    calls the API to get all the storage profiles.
    """
    deadline = get_boto3_client("deadline", config=config)

    return _call_paginated_deadline_list_api(
        deadline.list_storage_profiles_for_queue, "storageProfiles", **kwargs
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

__all__ = [
    "summarize_paths_by_nested_directory",
    "summarize_paths_by_sequence",
    "human_readable_file_size",
    "summarize_path_list",
    "PathSummary",
]

from ...common.path_utils import (
    human_readable_file_size,
    summarize_paths_by_nested_directory,
    summarize_paths_by_sequence,
    summarize_path_list,
    PathSummary,
)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

__all__ = [
    "login",
    "logout",
    "create_job_from_job_bundle",
    "wait_for_create_job_to_complete",
    "get_boto3_session",
    "get_boto3_client",
    "AwsAuthenticationStatus",
    "AwsCredentialsSource",
    "TelemetryClient",
    "check_authentication_status",
    "check_deadline_api_available",
    "get_credentials_source",
    "precache_clients",
    "list_farms",
    "list_queues",
    "list_jobs",
    "list_fleets",
    "list_storage_profiles_for_queue",
    "get_queue_user_boto3_session",
    "get_queue_parameter_definitions",
    "get_telemetry_client",
    "get_deadline_cloud_library_telemetry_client",
    "get_storage_profile_for_queue",
    "record_success_fail_telemetry_event",
    "record_function_latency_telemetry_event",
    "assume_queue_role_for_user",
    "assume_queue_role_for_read",
    "wait_for_job_completion",
    "JobCompletionResult",
    "FailedTask",
    "get_session_logs",
    "SessionLogResult",
    "LogEvent",
]

# The following import is needed to prevent the following sporadic failure:
# botocore.exceptions.HTTPClientError: An HTTP Client raised an unhandled exception: unknown
# encoding: idna
import encodings.idna  # noqa # pylint: disable=unused-import
from configparser import ConfigParser
from logging import getLogger
from typing import Any, Dict, Optional


# Telemetry must be imported before Submit Job Bundle to avoid circular dependencies.
from ._telemetry import (
    get_telemetry_client,
    get_deadline_cloud_library_telemetry_client,
    TelemetryClient,
    record_success_fail_telemetry_event,
    record_function_latency_telemetry_event,
)
from ._loginout import login, logout
from ._session import (
    AwsAuthenticationStatus,
    AwsCredentialsSource,
    precache_clients,
    get_queue_user_boto3_session,
    check_authentication_status,
    get_boto3_client,
    get_boto3_session,
    get_credentials_source,
    get_user_and_identity_store_id,
)
from ._list_apis import (
    list_farms,
    list_queues,
    list_jobs,
    list_fleets,
    list_storage_profiles_for_queue,
)
from ._queue_parameters import get_queue_parameter_definitions
from ._queue_credentials import (
    assume_queue_role_for_user,
    assume_queue_role_for_read,
)
from ._submit_job_bundle import (
    create_job_from_job_bundle,
    wait_for_create_job_to_complete,
)
from ._get_storage_profile_for_queue import get_storage_profile_for_queue
from ._job_monitoring import (
    wait_for_job_completion,
    JobCompletionResult,
    FailedTask,
    get_session_logs,
    SessionLogResult,
    LogEvent,
)

logger = getLogger(__name__)


def check_deadline_api_available(config: Optional[ConfigParser] = None) -> bool:
    """
    Returns True if AWS Deadline Cloud APIs are authorized in the session,
    False otherwise. This only checks the deadline:ListFarms API by performing
    one call with just one result.

    Args:
        config (ConfigParser, optional): The AWS Deadline Cloud configuration
                object to use instead of the config file.
    """
    import logging

    from ._session import _modified_logging_level

    with _modified_logging_level(logging.getLogger("botocore.credentials"), logging.ERROR):
        try:
            list_farm_params: Dict[str, Any] = {"maxResults": 1}
            user_id, _ = get_user_and_identity_store_id(config=config)
            if user_id:
                list_farm_params["principalId"] = str(user_id)

            deadline = get_boto3_client("deadline", config=config)
            deadline.list_farms(**list_farm_params)
            return True
        except Exception:
            logger.exception("Error invoking ListFarms")
            return False
